# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from .processor import Processor

def parse_residue_spec(resspec):
    """
    Light version of:
    vermouth.processors.annotate_mud_mod.parse_residue_spec

    Parse a residue specification:
    <mol_name>#<mol_idx>-<resname>#<resid>

    Returns a dictionary with keys 'mol_name',
    'mol_idx', and 'resname' for the fields that are specified.
    Resid will be an int.

    Parameters
    ----------
    resspec: str
    Returns
    -------
    dict
    """
    molname = None
    resname = None
    mol_idx = None
    resid = None
    mol_name_idx, *res = resspec.split('-', 1)
    molname, *mol_idx = mol_name_idx.split('#', 1)

    if res:
        resname, *resid = res[0].split('#', 1)

    error_msg = ('Your selection {} is invalid. Was not able to assign {}'
                 'with value {}')

    out = {}
    if mol_idx:

        try:
           mol_idx = int(mol_idx[0])
        except ValueError:
           raise IOError(error_msg.format(resspec, 'mol_idx', mol_idx))

        out['mol_idx'] = mol_idx

    if resname:
        out['resname'] = resname
    if molname:
        out['molname'] = molname
    if resid:
        try:
           out['resid'] = float(resid[0])
        except ValueError:
           raise IOError(error_msg.format(resspec, 'mol_idx', mol_idx))

    return out


def _find_nodes(molecule, mol_attr):
    """
    Find nodes with resname and resid
    attribute if they are defined in mol_attr.
    """
    attr = {}
    for key in ["resname", "resid"]:
        if key in mol_attr:
           attr[key] = mol_attr[key]

    for node in molecule.nodes:
        if all([molecule.nodes[node][key] == value for
                key, value in attr.items()]):

            yield node

class AnnotateLigands(Processor):
    """
    Given some molecules defined as Ligands,
    the processor adds them to the ligated molecules
    via an edge and indicates the node as ligated.
    This will cause it to be picked up by the build
    processor and place the ligands close. The processor
    also contains an extract method, which will readout
    the coordinates from the ligated molecules and puts
    them back such that the topology order is kept.
    """

    def __init__(self, topology, ligands):
        """
        Initalize the processor with a :class:`topology`
        and a list of `ligand definitions` which
        are used both in annotating and extracting
        the positions later. Format of ligand definitions
        needs to be in the format of:

        <mol_name>#<mol_idx>-<resname>#<resid>:<mol_name>#<mol_idx>-<resname>#<resid>

        Elements of this may be omitted. <resid> and <mol_idx> cannot
        contain - or # characters.
        """
        self.topology = topology
        self.ligand_defs = defaultdict(list)

        for mol_spec, lig_spec in ligands:
            mol_attr = parse_residue_spec(mol_spec)
            lig_attr = parse_residue_spec(lig_spec)

           # check which molecules are elligable for annotation
            if "mol_idx" in mol_attr:
                mol_idxs = [mol_attr["mol_idx"]]

                if "molname" in mol_attr:
                    mol_name = mol_attr["molname"]
                    allowed_idxs = self.topology.mol_idx_by_name[mol_name]
                    if any(idx not in allowed_idxs for idx in mol_idxs):
                       msg = ("Your molecule name {} does not"
                              " match your molecule index {}.")
                       raise IOError(msg.format(mol_name, mol_idxs))

            elif "molname" in mol_attr:
                mol_name = mol_attr["molname"]
                mol_idxs = self.topology.mol_idx_by_name[mol_name]
            # if neither molname nor index are provided all molecules
            # are eligible
            else:
                mol_idxs = range(0, len(self.topology.molecules))

            # check which ligands are elligable for annotation
            if "mol_idx" in lig_attr:
                lig_idxs = [lig_attr["mol_idx"]]
            elif "molname" in lig_attr:
                lig_name = lig_attr["molname"]
                lig_idxs = self.topology.mol_idx_by_name[lig_name]
            else:
                msg = ("Your ligand selection needs to conatin a molecule idx and/or "
                       "a molecule name. Currently the selection {} contains none.")
                raise IOError(msg.format(lig_spec))

            # make sure that each lig associated with a molecule
            # is a different molecule in the topology
            total = 0
            for mol_idx in mol_idxs:
                mol_nodes = _find_nodes(topology.molecules[mol_idx], mol_attr)
                for mol_node in mol_nodes:
                    self.ligand_defs[mol_idx].append((mol_node, lig_idxs[total], mol_attr, lig_attr))
                    total += 1

    def _connect_ligands_to_molecule(self, molecule, mol_idx):
        """
        Given a `molecule` check if any of the ligand
        specs correspond to that molecule and annotate the
        ligand if the specs match. A node is added to the
        matching molecule. This node as the additional
        attributes:

        ligated:   tuple(int, node_key)
            molecule index of the ligand, node_key of the ligands node
        build: True
            this node needs to be built
        """
        current = max(molecule.nodes) + 1
        for mol_node, lig_idx, mol_attr, lig_attr in self.ligand_defs[mol_idx]:
            ligand = self.topology.molecules[lig_idx]
            lig_nodes = _find_nodes(ligand, lig_attr)

            for lig_node in lig_nodes:
                resname = ligand.nodes[lig_node]["resname"]
                molecule.add_monomer(current, resname, [(mol_node, current)])

                molecule.nodes[current]["build"] = True
                molecule.nodes[current]["ligated"] = (lig_idx,
                                                      lig_node)
                current += 1

    def split_ligands(self):
        """
        Given the ligand specs extract ligand positions from the
        molecules and put them back in the inital topology molecules
        which discribe the ligands.
        """
        for molecule in self.topology.molecules:
            remove_nodes = []

            for node in molecule.nodes:
                if "ligated" in molecule.nodes[node]:
                    lig_idx, lig_node = molecule.nodes[node]["ligated"]
                    pos = molecule.nodes[node]["position"]
                    self.topology.molecules[lig_idx].nodes[lig_node]["position"] = pos
                    remove_nodes.append(node)

            for node in remove_nodes:
                molecule.remove_node(node)

    def run_molecule(self, meta_molecule, mol_idx):
        """
        Perform the random walk for a single molecule.
        """
        self._connect_ligands_to_molecule(meta_molecule, mol_idx)
        return meta_molecule

    def run_system(self, system):
        """
        Process `system`.
        Parameters
        ----------
        system: vermouth.system.System
            The system to process. Is modified in-place.
        """
        mols = []
        for mol_idx, molecule in enumerate(system.molecules):
            mols.append(self.run_molecule(molecule, mol_idx))
        system.molecules = mols



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""

import textwrap
import pytest
import numpy as np
import networkx as nx
import vermouth.forcefield
import vermouth.molecule
from polyply import TEST_DATA
import polyply.src.meta_molecule
from polyply.src.meta_molecule import (MetaMolecule, Monomer)
from .example_fixtures import example_meta_molecule

class TestPolyply:
    @staticmethod
    def test_add_monomer():
        ff = vermouth.forcefield.ForceField(name='test_ff')
        meta_mol = MetaMolecule(name="test", force_field=ff)
        meta_mol.add_monomer(0,"PEO",[])
        meta_mol.add_monomer(1,"PEO",[(1, 0)])

        assert nx.get_node_attributes(meta_mol, "resname") == {0: 'PEO', 1: 'PEO'}
        assert list(meta_mol.nodes) == [0, 1]
        assert list(meta_mol.edges) == [(0, 1)]

    @staticmethod
    def test_add_monomer_fail():
        ff = vermouth.forcefield.ForceField(name='test_ff')
        meta_mol = MetaMolecule(name="test", force_field=ff)
        meta_mol.add_monomer(0,"PEO",[])
        with pytest.raises(IOError):
            meta_mol.add_monomer(1,"PEO",[(1, 8)])

    @staticmethod
    def test_get_edge_resname():
        ff = vermouth.forcefield.ForceField(name='test_ff')
        meta_mol = MetaMolecule(name="test", force_field=ff)
        meta_mol.add_monomer(0,"PEO",[])
        meta_mol.add_monomer(1,"PEO",[(1, 0)])
        name = meta_mol.get_edge_resname((1, 0))
        assert name == ["PEO", "PEO"]

    @staticmethod
    @pytest.mark.parametrize('monomers, edges, nodes, attrs', (
        # multiple blocks from single monomer
          ([Monomer(resname="PEO", n_blocks=2)],
           [(0,1)],
           [0,1],
           {0: 'PEO', 1: 'PEO'}
          ),
        # two blocks from two monomers
          ([Monomer(resname="PEO", n_blocks=1), Monomer(resname="PS",n_blocks=1)],
           [(0,1)],
           [0,1],
           {0: 'PEO', 1: 'PS'}
           ),
        # multiple blocks from two monomers
          ([Monomer(resname="PEO", n_blocks=2), Monomer(resname="PS",n_blocks=2)],
           [(0,1),(1,2),(2,3)],
           [0,1,2,3],
           {0: 'PEO', 1: 'PEO', 2: 'PS', 3: 'PS'}
         )))
    def test_from_monomer_seq_linear(monomers, edges, nodes, attrs):
        ff = vermouth.forcefield.ForceField(name='test_ff')
        name = "test"
        meta_mol = MetaMolecule.from_monomer_seq_linear(ff, monomers, name)

        assert nx.get_node_attributes(meta_mol, "resname") == attrs
        assert list(meta_mol.nodes) == nodes
        assert list(meta_mol.edges) == edges


    @staticmethod
    @pytest.mark.parametrize('file_name, edges, nodes, attrs', (
        # multiple blocks from single monomer
          (TEST_DATA + "/json/linear.json",
           [(0,1), (1,2)],
           [0,1,2],
           {0: 'PEO', 1: 'PEO', 2: 'PEO'}
          ),
       # two blocks from two monomers
         (TEST_DATA + "/json/single_branch.json",
          [(1,2),(2,3),(3,4),(2,5)],
          [1,2,3,4,5],
          {1: 'PEO', 2: 'PEO', 3: 'PS', 4: 'PS', 5:'PEO'}
          ),
       # two blocks from two monomers
         (TEST_DATA + "/json/double_branch.json",
          [(1,2),(2,3),(2,4),(4,5),(5,6),(5,8),(6,7)],
          [1,2,3,4,5,6,7,8],
          {1: 'PEO', 2: 'PEO', 3: 'PS', 4: 'PEO', 5:'PEO', 6: 'PS',
           7: 'PS', 8: 'PEO'}
          ),
       # Hyperbranched
         (TEST_DATA + "/json/hyperbranched.json",
          [(0, 1), (0, 2), (0, 3), (1, 4), (1, 5), (2, 6), (2, 7), (3, 8), (3, 9),
           (4, 10), (4, 11), (5, 12), (5, 13)],
          [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13],
          {0: 'N1', 1: 'N2', 2: 'N3', 3: 'N3', 4: 'N2', 5: 'N3', 6: 'N3',
           7: 'N1', 8: 'N2', 9: 'N3', 10: 'N3', 11: 'N2', 12: 'N3', 13: 'N3'}
          ),
        # check that ordering is restored
          (TEST_DATA + "/json/linear_rev.json",
           [(0, 1), (1, 2)],
           [0, 1, 2],
           {0: 'PEO', 1: 'PEO', 2: 'PEO'}
          ),
           ))
    def test_from_json(file_name, edges, nodes, attrs):
        ff = vermouth.forcefield.ForceField(name='test_ff')
        name = "test"
        meta_mol = MetaMolecule.from_json(ff, file_name, name)

        assert len(nx.get_node_attributes(meta_mol, "resid")) == len(nodes)
        #assert nx.get_node_attributes(meta_mol, "resname") == attrs
        print(meta_mol.edges)
        assert set(meta_mol.nodes) == set(nodes)
        assert set(meta_mol.edges) == set(edges)

    @staticmethod
    def test_from_json_error_resid():
        ff = vermouth.forcefield.ForceField(name='test_ff')
        name = "test"
        file_name = TEST_DATA + "/json/fail.json"
        with pytest.raises(IOError):
            meta_mol = MetaMolecule.from_json(ff, file_name, name)

    @staticmethod
    def test_from_itp():
       file_name = TEST_DATA + "/itp/PEO.itp"
       edges = [(0,1), (1,2)]
       nodes = [0, 1, 2]
       attrs = {0: 'PEO', 1: 'PEO', 2: 'PEO'}

       ff = vermouth.forcefield.ForceField(name='test_ff')
       name = "PEO"
       meta_mol = MetaMolecule.from_itp(ff, file_name, name)

       assert set(meta_mol.nodes) == set(nodes)
       assert set(meta_mol.edges) == set(edges)

@pytest.mark.parametrize('split_pattern, expct_high_res, expct_low_res, edges, resid_low, resid_high', (
    # split single residues of one type into two
    (
    ["B:B1-BB,BB1,BB2:B2-SC1,SC2"],
    {0: "A", 1: "A", 2: "A", 3: "A", 4: "B1", 5: "B1", 6: "B1", 7: "B2", 8: "B2",
     9: "A", 10: "A", 11: "A", 12: "A"},
    {0: "A", 1: "B1", 2: "B2", 3: "A"},
    [(0, 1), (1, 3), (1, 2)],
    {0: 0, 1: 1, 2: 2, 3: 3},
    {0: 0, 1: 0, 2: 0, 3: 0, 4: 1, 5: 1, 6: 1, 7: 2, 8: 2,
     9: 3, 10: 3, 11: 3, 12: 3},
    ),
    # split two residues of one type into two
    (
    ["A:A1-BB,BB1:A2-SC1,SC2"],
    {0: "A1", 1: "A1", 2: "A2", 3: "A2", 4: "B", 5: "B", 6: "B", 7: "B", 8: "B",
     9: "A1", 10: "A1", 11: "A2", 12: "A2"},
    {0: "A1", 1: "A2", 2: "B", 3: "A1", 4: "A2"},
    [(0, 1), (0, 2), (2, 3), (3, 4)],
    {0: 0, 1: 1, 2: 2, 3: 3, 4: 4},
    {0: 0, 1: 0, 2: 1, 3: 1, 4: 2, 5: 2, 6: 2, 7: 2, 8: 2,
     9: 3, 10: 3, 11: 4, 12: 4},
    ),
    # split all residues
    (
    ["A:A1-BB,BB1:A2-SC1,SC2", "B:B1-BB,BB1,BB2:B2-SC1,SC2"],
    {0: "A1", 1: "A1", 2: "A2", 3: "A2", 4: "B1", 5: "B1", 6: "B1", 7: "B2", 8: "B2",
     9: "A1", 10: "A1", 11: "A2", 12: "A2"},
    {0: "A1", 1: "A2", 2: "B1", 3: "B2", 4: "A1", 5: "A2"},
    [(0, 1), (0, 2), (2, 3), (2, 4), (4, 5)],
    {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5},
    {0: 0, 1: 0, 2: 1, 3: 1, 4: 2, 5: 2, 6: 2, 7: 3, 8: 3,
     9: 4, 10: 4, 11: 5, 12: 5},
    ),
    ))
def test_split_residue(example_meta_molecule,
                       split_pattern,
                       expct_high_res,
                       expct_low_res,
                       edges,
                       resid_low,
                       resid_high):

    example_meta_molecule.split_residue(split_pattern)
    new_resnames_high = nx.get_node_attributes(example_meta_molecule.molecule, "resname")
    assert new_resnames_high == expct_high_res
    new_resnames_low = nx.get_node_attributes(example_meta_molecule, "resname")
    assert new_resnames_low == expct_low_res
    new_resid_low = nx.get_node_attributes(example_meta_molecule, "resid")
    assert new_resid_low == resid_low
    new_resid_high = nx.get_node_attributes(example_meta_molecule.molecule, "resid")
    assert new_resid_high == resid_high

    for node_a, node_b in edges:
        assert example_meta_molecule.has_edge(node_a, node_b)

def test_split_residue_err(example_meta_molecule):
    split_pattern = ["A:A1-BB,BB1:A2-BB1,SC1,SC2"]
    with pytest.raises(IOError):
         example_meta_molecule.split_residue(split_pattern)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
High level API for the polyply coordinate generator
"""
import sys
from collections import defaultdict
from functools import partial
import multiprocessing
import numpy as np
import networkx as nx
from tqdm import tqdm
import vermouth.forcefield
from vermouth.file_writer import DeferredFileWriter
from vermouth.log_helpers import StyleAdapter, get_logger
from .generate_templates import GenerateTemplates
from .backmap import Backmap
from .topology import Topology
from .build_system import BuildSystem
from .annotate_ligands import AnnotateLigands, parse_residue_spec, _find_nodes
from .build_file_parser import read_build_file

LOGGER = StyleAdapter(get_logger(__name__))

def find_starting_node_from_spec(topology, start_nodes):
    """
    Given a definition of one or multiple nodes in
    `start_nodes` find the corresponding molecule index
    and node key in `topology`. The format of start nodes
    is equivalent to that for annotating the residue and
    is described in detail in the function
    :meth:`polyply.src.annotate_ligands.annotate_molecules.parse_residue_spec`.

    Parameters:
    -----------
    topology: :class:`polyply.src.topology.Topology`
    start_nodes: list[str]
        a list of residue-specs

    Returns:
    --------
    dict
        a dict of molecule_idx as key and node key as value
    """
    start_dict = {mol_idx: None for mol_idx in range(len(topology.molecules))}
    for start in start_nodes:
        res_spec = parse_residue_spec(start)
        if 'mol_idx' in res_spec:
            mol_idx = res_spec['mol_idx']
            node = list(_find_nodes(topology.molecules[mol_idx], res_spec))[0]
            start_dict[mol_idx] = node
        else:
            for idx, molecule in enumerate(topology.molecules):
                if molecule.mol_name == res_spec['molname']:
                    node = list(_find_nodes(molecule, res_spec))[0]
                    start_dict[idx] = node
    return start_dict

def _check_molecules(molecules):
    """
    Helper method which raises an IOError
    if any molecule in `molecules` is
    disconnected.
    """
    # check if molecules are all connected
    for molecule in molecules:
        if not nx.is_connected(molecule):
            msg = ('\n Molecule {} consistes of two disconnected parts. '
                   'Make sure all atoms/particles in a molecule are '
                   'connected by bonds, constraints or virual-sites')
            raise IOError(msg.format(molecule.name))

def gen_coords(args):
    # Read in the topology
    LOGGER.info("reading topology",  type="step")
    topology = Topology.from_gmx_topfile(name=args.name, path=args.toppath)

    LOGGER.info("processing topology",  type="step")
    topology.preprocess()
    _check_molecules(topology.molecules)

    if args.split:
       LOGGER.info("splitting residues",  type="step")
       for molecule in topology.molecules:
           molecule.split_residue(args.split)

    # read in coordinates if there are any
    if args.coordpath:
        LOGGER.info("loading coordinates",  type="step")
        topology.add_positions_from_file(args.coordpath, args.build_res)

    # load in built file
    if args.build:
        LOGGER.info("reading build file",  type="step")
        with open(args.build) as build_file:
            lines = build_file.readlines()
            read_build_file(lines, topology.molecules)

    # collect all starting points for the molecules
    start_dict = find_starting_node_from_spec(topology, args.start)

    # handle grid input
    if args.grid:
        LOGGER.info("loading grid",  type="step")
        args.grid = np.loadtxt(args.grid)

    # Build polymer structure
    LOGGER.info("generating templates",  type="step")
    GenerateTemplates(topology=topology, max_opt=10).run_system(topology)
    LOGGER.info("annotating ligands",  type="step")
    AnnotateLigands(topology, args.ligands).run_system(topology)
    LOGGER.info("generating system coordinates",  type="step")
    BuildSystem(topology,
                start_dict=start_dict,
                density=args.density,
                max_force=args.max_force,
                grid_spacing=args.grid_spacing,
                maxiter=args.maxiter,
                box=args.box,
                step_fudge=args.step_fudge,
                ignore=args.ignore,
                grid=args.grid,
                nrewind=args.nrewind).run_system(topology.molecules)
    AnnotateLigands(topology, args.ligands).split_ligands()
    LOGGER.info("backmapping to target resolution",  type="step")
    Backmap(fudge_coords=args.bfudge).run_system(topology)
    # Write output
    LOGGER.info("writing output",  type="step")
    command = ' '.join(sys.argv)
    system = topology.convert_to_vermouth_system()
    vermouth.gmx.gro.write_gro(system, args.outpath, precision=7,
                               title=command, box=topology.box)
    DeferredFileWriter().write()



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""
import pytest
import textwrap
import numpy as np
from vermouth.forcefield import ForceField
import polyply
from polyply.src.topology import Topology
from polyply.src.top_parser import read_topology
from polyply.src.build_system import (_compute_box_size,
                                      _filter_by_molname,
                                      BuildSystem
                                     )

@pytest.fixture
def example_topology():
    top_lines ="""
    [ defaults ]
    1   1   no   1.0     1.0
    [ atomtypes ]
    N0 45.0 0.000 A 0.0 0.0
    [ nonbond_params ]
    N0   N0   1 4.700000e-01    3.700000e+00
    [ moleculetype ]
    testA 1
    [ atoms ]
    1    N0  1   GLY    BB   1 0.00     45
    2    N0  1   GLY    SC1  1 0.00     45
    3    N0  1   GLY    SC2  1 0.00     45
    4    N0  2   GLU    BB   2 0.00     45
    5    N0  2   GLU    SC1  2 0.00     45
    6    N0  2   GLU    SC2  2 0.00     45
    [ bonds ]
    1    2    1  0.47 2000
    2    3    1  0.47 2000
    1    4    1  0.47 2000
    4    5    1  0.47 2000
    4    6    1  0.47 2000
    [ moleculetype ]
    testB 1
    [ atoms ]
    1    N0  1   ASP    BB   1 0.00
    2    N0  1   ASP    SC1  1 0.00
    3    N0  1   ASP    SC2  1 0.00
    [ bonds ]
    1    2    1  0.47 2000
    2    3    1  0.47 2000
    [ system ]
    test system
    [ molecules ]
    testA 2
    testB 1
    """
    lines = textwrap.dedent(top_lines)
    lines = lines.splitlines()
    force_field = ForceField("test")
    topology = Topology(force_field)
    read_topology(lines=lines, topology=topology, cwdir="./")
    topology.preprocess()
    topology.volumes = {"GLY": 0.53, "GLU": 0.67, "ASP": 0.43}
    return topology

@pytest.mark.parametrize('density, result', (
    (1000.0,
     1.0387661572640916
    ),
    (600.0,
     1.2315934632345062
     )))
def test_compute_box_size(example_topology, density, result):
    top = example_topology
    assert np.isclose(_compute_box_size(top, density), result)


def test_compute_box_size_error(example_topology):
    top = example_topology
    # clear the atomtypes to see if error is raised
    # when no mass can be found
    top.atom_types = {}
    with pytest.raises(KeyError):
        _compute_box_size(top, 1000)

@pytest.mark.parametrize('ignore', (
    ["test_A"],
    ["test_A", "test_B"],
     ))
def test_filer_by_molname(example_topology, ignore):
    molecules = _filter_by_molname(example_topology.molecules, ignore)
    for molecule in molecules:
        assert molecule.mol_name not in ignore


@pytest.mark.parametrize('box, density, grid_spacing, grid, expected_grid_shape, expected_box', (
   # grid and box are generated automatically
   (None,
    1000,
    0.2,
    None,
    (216, 3),
    np.array([1.03877, 1.03877, 1.03877])
   ),
   # box is provided and grid is generated automatically
   (np.array([5.0, 5.0, 5.0]),
    1000,
    0.2,
    None,
    (15625, 3),
    np.array([5.0, 5.0, 5.0]),
   ),
   # box is provided and grid also nothing is generated
   (np.array([5.0, 5.0, 5.0]),
    1000,
    0.2,
    np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
    (2, 3),
    np.array([5.0, 5.0, 5.0]),
   ),
   ))
def test_init(example_topology,
             box, density,
             grid_spacing, grid,
             expected_grid_shape,
             expected_box):

   processor = BuildSystem(example_topology,
                           start_dict={},
                           box=box,
                           density=density,
                           grid=grid,
                           grid_spacing=grid_spacing,
                           )
   assert processor.box_grid.shape == expected_grid_shape
   assert all(processor.box == expected_box)

def test_rw_arg_failure():
    with pytest.raises(TypeError):
         BuildSystem(example_topology,
                     density=1000,
                     start_dict={},
                     random_argument=10)

@pytest.mark.parametrize("""positions, box,
                          starting_grid,
                          starting_nodes,
                          """, (
   # no inital starting positoins
   ([],
    np.array([5., 5., 5.]),
    np.array([[2.0, 2.0, 2.0],
              [3.0, 3.0, 3.0],
              [0.5, 0.5, 0.5],
              [4.0, 4.0, 4.0]]),
    {0:0, 1:0, 2:0}
   ),
   # skip first molecule
   (np.array([[1.25, 1.00, 1.00],
              [1.75, 1.00, 1.00],
             ]),
    np.array([5., 5., 5.]),
    np.array([[2.0, 2.0, 2.0],
              [3.0, 3.0, 3.0],
              [0.5, 0.5, 0.5],
              [4.0, 4.0, 4.0]]),
   {0:0, 1:0, 2:0}
   ),
   # partial starting positions
   (np.array([[1.25, 1.00, 1.00]]),
    np.array([5., 5., 5.]),
    np.array([[2.0, 2.0, 2.0],
              [3.0, 3.0, 3.0],
              [0.5, 0.5, 0.5],
              [4.0, 4.0, 4.0]]),
    {0:0, 1:0, 2:0}
   ),
   # scramble starting nodes
   ([],
    np.array([5., 5., 5.]),
    np.array([[2.0, 2.0, 2.0],
              [3.0, 3.0, 3.0],
              [0.5, 0.5, 0.5],
              [4.0, 4.0, 4.0]]),
    {0:1, 1:0, 2:0}
   ),
   ))
def test_build_system(positions,
                      example_topology,
                      box,
                      starting_grid,
                      starting_nodes):

   total = 0
   for mol in example_topology.molecules:
       for node in mol.nodes:
           try:
               mol.nodes[node]["position"] = positions[total]
               mol.nodes[node]["build"] = False
           except IndexError:
               mol.nodes[node]["build"] = True
           total += 1


   processor = BuildSystem(example_topology,
                           density=1000,
                           start_dict=starting_nodes,
                           box=box,
                           grid=starting_grid,
                           )
   processor.run_system(example_topology)
   total = 0
   for mol_idx, mol in enumerate(example_topology.molecules):
       for node in mol.nodes:
           if total < len(positions):
              assert all(mol.nodes[node]["position"] == positions[total])
           elif node == starting_nodes[mol_idx]:
              assert mol.nodes[node]["position"] in starting_grid
           else:
              assert "position" in mol.nodes[node]
           total += 1



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import networkx as nx
from networkx.readwrite import json_graph
import pytest
from polyply import gen_seq, TEST_DATA
from polyply.src.gen_seq import (_add_edges, _branched_graph, _tag_nodes,
                                 MacroString, generate_seq_graph,
                                 _find_terminal_nodes, _apply_termini_modifications)


def test_add_edge():
    graph = nx.Graph()
    graph.add_nodes_from(list(range(0, 6)))
    graph.add_edges_from([(0, 1), (1, 2),
                          (3, 4), (4, 5)])
    nx.set_node_attributes(graph,
                           {0: 1, 1: 1, 2: 1, 3: 2, 4: 2, 5: 2},
                           "seqid")
    edges = "1-2"
    _add_edges(graph, edges, 1, 2)
    assert graph.has_edge(1, 5)


@pytest.mark.parametrize("branching_f, n_levels, ref_edges", (
    (2, 3, [(0, 1), (0, 2), (1, 3), (1, 4), (2, 5), (2, 6)]),
    (1, 4, [(0, 1), (1, 2), (2, 3)])
))
def test_branched_graph(branching_f, n_levels, ref_edges):
    graph = _branched_graph("test", branching_f, n_levels)
    resnames = {idx: "test" for idx in graph.nodes}
    assert len(graph.edges) == len(ref_edges)
    for edge in ref_edges:
        graph.has_edge(edge[0], edge[1])
    assert nx.get_node_attributes(graph, "resname") == resnames


@pytest.mark.parametrize("macro_str, ref_edges, seed, resnames", (
    ("A:3:1:PEO-1.", [(0, 1), (1, 2)], None,
     {0: "PEO", 1: "PEO", 2: "PEO"}),
    ("A:3:2:PPI-1.",
     [(0, 1), (0, 2), (1, 3), (1, 4), (2, 5), (2, 6)],
     None,
     {0:"PPI", 1:"PPI", 2:"PPI", 3:"PPI", 4:"PPI", 5:"PPI", 6:"PPI"}),
    ("A:4:1:PEO-0.5,PPO-0.5",
     [(0, 1), (1, 2), (2, 3)],
     69034,
     {0:"PPO", 1:"PPO", 2:"PPO", 3:"PEO"}),
    ("A:4:1:PEO-0.3,PPO-0.2",
     [(0, 1), (1, 2), (2, 3)],
     69034,
     {0:"PPO", 1:"PPO", 2:"PPO", 3:"PEO"})
))
def test_interpret_macro_string(macro_str, ref_edges, seed, resnames):
    ref_graph = nx.Graph()
    ref_graph.add_edges_from(ref_edges)
    macro = MacroString(macro_str)
    macro_graph = macro.gen_graph(seed=seed)
    resnames_macro = nx.get_node_attributes(macro_graph, "resname")
    assert resnames == resnames_macro
    assert ref_graph.edges == macro_graph.edges


def test_generate_seq_graph():
    macro_str_A = "A:6:1:A-1."
    macro_str_B = "B:2:4:B-1."

    macros = {"A": MacroString(macro_str_A),
              "B": MacroString(macro_str_B)}

    seq_graph = generate_seq_graph(
        ["A", "B", "A"], macros, ["0:1:2-0", "1:2:1-1"])
    graphA = macros["A"].gen_graph()
    graphB = macros["B"].gen_graph()

    ref_graph = nx.disjoint_union(graphA, graphB)
    ref_graph = nx.disjoint_union(ref_graph, graphA)
    ref_graph.add_edges_from([(2, 6), (7, 12)])
    assert nx.is_isomorphic(ref_graph, seq_graph)


@pytest.mark.parametrize("edge_str",
                         ("0:1:2-100",
                          "0:100:2-0"))
def test_generate_seq_edge_error(edge_str):
    macro_strA = "A:6:1:A-1."
    macro_strB = "B:2:4:B-1."

    macros = {"A": MacroString(macro_strA),
              "B": MacroString(macro_strB)}

    with pytest.raises(IOError):
        generate_seq_graph(["A", "B"], macros, [edge_str])


def test_find_termini():
    graph = nx.Graph()
    graph.add_edges_from([(0, 1), (1, 2), (1, 3), (3, 4), (3, 5)])
    termini = _find_terminal_nodes(graph)
    for node in [0, 2, 4, 5]:
        assert node in termini


def test_annote_modifications():
    modf = ["1:NH2", "2:NH3"]
    graph = nx.Graph()
    graph.add_edges_from([(0, 1), (1, 2), (1, 3), (3, 4), (3, 5)])
    nx.set_node_attributes(graph,
                           {0: 1, 1: 1, 2: 1, 3: 2, 4: 2, 5: 2},
                           "seqid"
                          )
    nx.set_node_attributes(graph,
                           {0: "PPI", 1: "PPI", 2: "PPI", 3: "PPI", 4: "PPI", 5: "PPI"},
                           "resname"
                          )
    _apply_termini_modifications(graph, modf)
    assert nx.get_node_attributes(graph, "resname") == {0: "NH2", 1: "PPI",
                                                        2: "NH2", 3: "PPI",
                                                        4: "NH3", 5: "NH3"}

#<seqID:tag_word:value1-probability,value2-probability>
@pytest.mark.parametrize('tags, expected, seed',(
                         # only one sequence, one value, prob=1
                        [["1:chiral:R-1."],
                         {0: "R", 1: "R", 2: "R"},
                         None
                        ],
                         # two tags different seqID value prob=1
                        [["1:chiral:R-1.", "2:chiral:S-1."],
                         {0: "R", 1: "R", 2: "R", 3: "S", 4: "S", 5: "S"},
                         None
                        ],
                        # randomly replace nodes in one sequence
                        [["1:chiral:R-0.5,S-0.5", "2:chiral:Q-1."],
                         {0: "S", 1: "R", 2: "S", 3: "Q", 4: "Q", 5: "Q"},
                         23303
                        ],
                        # randomly replace nodes in both sequences
                        [["1:chiral:R-0.5,S-0.5", "2:chiral:P-0.5,D-0.5"],
                         {0: "S", 1: "R", 2: "S", 3: "D", 4: "P", 5: "D"},
                         23303
                        ]
                        ))
def test_tag_nodes(tags, expected, seed):
    graph = nx.Graph()
    graph.add_edges_from([(0, 1), (1, 2), (1, 3), (3, 4), (3, 5)])
    nx.set_node_attributes(graph,
                           {0: 1, 1: 1, 2: 1, 3: 2, 4: 2, 5: 2},
                           "seqid"
                          )
    nx.set_node_attributes(graph,
                           {0: "PPI", 1: "PPI", 2: "PPI", 3: "PPI", 4: "PPI", 5: "PPI"},
                           "resname"
                          )
    _tag_nodes(graph, tags, seed)
    assert nx.get_node_attributes(graph, "chiral") == expected

class Args:
    """
    Inpute Arguments for the sequence generator.
    """

    def __init__(self,
                 name=None,
                 ffpath=None,
                 outpath=None,
                 macros=None,
                 file_macros=None,
                 seq=None,
                 connects=None,
                 lib=None,
                 modifications=[],
                 inpath=None,
                 tags=[]):

        self.name = name
        self.lib = lib
        self.ffpath = ffpath
        self.outpath = outpath
        self.macros = macros
        self.from_file = file_macros
        self.seq = seq
        self.connects = connects
        self.inpath = inpath
        self.modifications = modifications
        self.tags = []

@pytest.mark.parametrize('_input, ref_file', (
    (dict(outpath=TEST_DATA + "/gen_seq/output/PPI.json",
          macros=["A:3:2:N1-1.0"],
          seq=["A", "A"],
          connects=["0:1:0-0"]),
     TEST_DATA + "/gen_seq/ref/PPI_ref.json"),
    (dict(outpath=TEST_DATA + "/gen_seq/output/PEO_PS.json",
          macros=["A:11:1:PEO-1", "B:11:1:PS-1"],
          connects=["0:1:10-0"],
          seq=["A", "B"]),
     TEST_DATA + "/gen_seq/ref/PEO_PS_ref.json"),
    (dict(outpath=TEST_DATA + "/gen_seq/output/lysoPEG.json",
          inpath=[TEST_DATA + "/gen_seq/input/molecule_0.itp"],
          macros=["A:5:1:PEG-1.0"],
          file_macros=["PROT:molecule_0"],
          seq=["PROT", "A"],
          connects=["0:1:0-0"]),
     TEST_DATA + "/gen_seq/ref/lyso_PEG.json")
))
def test_gen_seq(_input, ref_file):
    arguments = Args(**_input)
    gen_seq(arguments)

    with open(ref_file) as _file:
        js_graph = json.load(_file)
        ref_graph = json_graph.node_link_graph(js_graph)

    with open(_input["outpath"]) as _file:
        js_graph = json.load(_file)
        out_graph = json_graph.node_link_graph(js_graph)

    assert nx.is_isomorphic(out_graph, ref_graph)
    assert nx.get_node_attributes(
        out_graph, "resname") == nx.get_node_attributes(ref_graph, "resname")



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
from vermouth.parser_utils import (
    SectionLineParser, _tokenize, _substitute_macros, _parse_macro
)
from vermouth.ffinput import FFDirector, _get_atoms, _treat_atom_prefix

class PolyplyFFParser(FFDirector):
    '''
    More strictly implements differences between .ff within polyply
    and .ff within martinize2.
    '''
    @SectionLineParser.section_parser('moleculetype', 'edges',
                                      negate=False, context_type='block')
    @SectionLineParser.section_parser('moleculetype', 'non-edges',
                                      negate=True, context_type='block')
    @SectionLineParser.section_parser('link', 'edges',
                                      negate=False, context_type='link')
    @SectionLineParser.section_parser('link', 'non-edges',
                                      negate=True, context_type='link')
    @SectionLineParser.section_parser('modification', 'edges',
                                      negate=False, context_type='modification')

    def _edges(self, line, lineno=0, negate=False, context_type=''):
        context = self.get_context(context_type)
        tokens = collections.deque(_tokenize(line))
        _parse_edges_new(tokens, context, context_type, negate=negate)

def _parse_edges_new(tokens, context, context_type, negate):
    """
    Parse the edge directive and add edge-attributes when
    required.
    """
    if negate:
        raise IOError("Polyply does not support non-edges.")

    atoms = _get_atoms(tokens, natoms=2)
    prefixed_atoms = []
    for idx, atom in enumerate(atoms):
        prefixed_reference, attributes = _treat_atom_prefix(*atom)
        for key in ["atomname", "order", "resname"]:
            attributes.pop(key, None)
        prefixed_atoms.append(prefixed_reference)

        if idx == 0 and attributes:
            raise IOError("The edge directive does not allow to set node attributes.")
        else:
            edge_attributes = attributes

    error_message = 'Atom with name {} not found for {} {}'
    for prefixed_atom in prefixed_atoms:
        atomname = prefixed_atom[0]
        if atomname not in context and context_type == 'modification':
            raise KeyError(error_message.format(atomname, context_type,
                                                context.name))
    context.add_edge(prefixed_atoms[0], prefixed_atoms[1], **edge_attributes)

def read_ff(lines, force_field):
    director = PolyplyFFParser(force_field)
    return list(director.parse(iter(lines)))



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import json
import networkx as nx
from networkx.readwrite import json_graph
from vermouth.graph_utils import make_residue_graph
from .load_library import load_library
from .generate_templates import find_atoms


def _branched_graph(resname, branching_f, n_levels):
    """
    Generate a tree graph of branching factor `branching_f`
    and number of generations as `n_levels` - 1. Each node
    gets the attribute resname.

    Parameters
    ----------
    resname:  str
    branching_f: int
    n_levels: int

    Returns
    -------
    `:class:networkx.Graph`
    """
    graph = nx.balanced_tree(r=branching_f, h=n_levels-1)
    resnames = {idx: resname for idx in graph.nodes}
    nx.set_node_attributes(graph, resnames, "resname")
    return graph


def _random_replace_nodes_attribute(graph, attribute_values,
                                    weights, attribute,
                                    nodes=[], seed=None):
    """
    Randomly replace or set attribute of `graph`
    based on the names in `attributes_values` taking into
    account `weights` provided. If nodes are provided
    they must be in graph.

    Parameters
    ----------
    graph: `:class:networkx.Graph`
    attribute_values: list[str]
    weights: list[float]
    attribute: str
        the attribute to be replaced
    seed: float
    nodes: list
        list of ellegible nodes

    Returns:
    --------
    `:class:networkx.Graph`
    """
    if not nodes:
        nodes = list(graph.nodes)

    random.seed(seed)
    for node in nodes:
        value = random.choices(attribute_values, weights=weights)
        graph.nodes[node][attribute] = value[0]

    return graph


class MacroString():
    """
    Define a (random) tree graph based on a string.
    The graph is generated each time the `gen_graph`
    attribute is run. For random graph they will be
    independent.
    """

    def __init__(self, string):
        """
        Convert string into definitions of the
        random tree graph.

        Parameters:
        -----------
        string: str
           the input string containing all definitions
        """
        input_params = string.split(":")
        self.name = input_params[0]
        self.levels = int(input_params[1])
        self.bfact = int(input_params[2])
        self.res_probs = input_params[3].split(',')

        self.residues = []
        self.weights = []
        for res_prob in self.res_probs:
            name, prob = res_prob.split("-")
            self.residues.append(name)
            self.weights.append(float(prob))

    def gen_graph(self, seed=None):
        """
        Generate a graph from the definitions stored in this
        instance a random seed can be provided.
        """
        graph = _branched_graph("dum", self.bfact, self.levels)
        graph = _random_replace_nodes_attribute(graph, self.residues,
                                                self.weights, "resname",
                                                seed=seed)
        return graph

class MacroFile():
    """
    Define a macro as residue graph of a
    vermouth molecule.
    """

    def __init__(self, name, force_field):
        """
        Select the appropiate molecule from
        the force_field.

        Parameters:
        -----------
        name: str
        force_field: `:class:vermouth.froce_field.ForceField`
        """
        self.molecule = force_field.blocks[name]

    def gen_graph(self, seed=None):
        """
        Generate a graph from the definitions stored in this
        instance by converting a molecule to a residue graph
        and setting the resname attribute.
        """
        block = make_residue_graph(self.molecule, attrs=('resid', 'resname'))
        resnames = nx.get_node_attributes(block, 'resname')
        graph = nx.Graph()
        graph.add_nodes_from(block.nodes)
        graph.add_edges_from(block.edges)
        nx.set_node_attributes(graph, resnames, "resname")
        return graph


def _add_edges(graph, edges, idx, jdx):
    """
    Add edges to a graph using the edge format
    'node_1,node_2', where the node keys refer
    to the node relative to the nodes with the same
    attribute `seqid`. The sequence ids are provided by
    `idx` and `jdx`.

    Parameters
    ----------
    graph: `:class:networkx.Graph`
    edges:  list[str]
        str has the format `node_keyA-node_keyB,node_keyC-node_keyD`,
        where edges will be added between nodes A and B, and C and D
    idx: int
    jdx: int

    Returns
    --------
    `:class:networkx.Graph`
    """
    for edge in edges.split(","):
        node_idx, node_jdx = map(str.strip, edge.split('-'))
        idx_nodes = find_atoms(graph, "seqid", idx)
        jdx_nodes = find_atoms(graph, "seqid", jdx)

        if not idx_nodes:
            msg = ("Trying to add connect between block with seqid {} and block with"
                   "seqid {}. However, cannot find block with seqid {}.")
            raise IOError(msg.format(idx, jdx, idx))
        elif not jdx_nodes:
            msg = ("Trying to add connect between block with seqid {} and block with"
                   "seqid {}. However, cannot find block with seqid {}.")
            raise IOError(msg.format(idx, jdx, jdx))

        try:
            node_i = idx_nodes[int(node_idx)]
        except IndexError:
            msg = ("Trying to add connect between block with seqid {} and block with"
                   "seqid {}. However, cannot find resid {} in block with seqid {}.")
            raise IOError(msg.format(idx, jdx, node_idx, idx))

        try:
            node_j = jdx_nodes[int(node_jdx)]
        except IndexError:
            msg = ("Trying to add connect between block with seqid {} and block with"
                   "seqid {}. However, cannot find resid {} in block with seqid {}.")
            raise IOError(msg.format(idx, jdx, node_idx, idx))

        graph.add_edge(node_i, node_j)

    return graph


def generate_seq_graph(sequence, macros, connects):
    """
    Given a sequence definition in terms of blocks in macros
    generate a graph and add the edges provided in connects.

    Parameters:
    ----------
    sequence: list
      a list of blocks defining a sequence
    macros: dict[str, networkx.graph]
      a dictionary of graphs defining a macro block
    connects: list[str]
      a list of connects

    Returns:
    -------
    `:class:networkx.graph`
    """
    seq_graph = nx.Graph()
    for idx, macro_name in enumerate(sequence):
        sub_graph = macros[macro_name].gen_graph()

        nx.set_node_attributes(
            sub_graph, {node: idx for node in sub_graph.nodes}, "seqid")
        seq_graph = nx.disjoint_union(seq_graph, sub_graph)

    for connect in connects:
        idx, jdx, edges = connect.split(":")
        _add_edges(seq_graph, edges, int(idx), int(jdx))

    return seq_graph


def _find_terminal_nodes(graph):
    """
    Find all termini of a graph and return the node key.
    A termini is defined as a node with degree of 1. Note
    that graph is assumed to be undirected.
    """
    termini = []
    for node in graph.nodes:
        if graph.degree(node) == 1:
            termini.append(node)

    return termini

def _apply_termini_modifications(graph, modifications):
    """
    Given a `graph` change the resname attribute of the end
    nodes as specified in modifications.

    Parameters:
    -----------
    graph: nx.Graph
    modifications: list[str:str]
        list of seq_ID:resname, where resname is the name
        to be applied to all terminal nodes part of block
        with seq_ID
    """
    terminal_nodes = _find_terminal_nodes(graph)
    for modification in modifications:
        seq_ID, resname = modification.split(':')
        idx_nodes = find_atoms(graph, "seqid", int(seq_ID))
        for node in idx_nodes:
            if node in terminal_nodes:
                graph.nodes[node]["resname"] = resname

def _tag_nodes(graph, tags, seed=None):
    """
    Find and label nodes as specified in a list of tags.
    Tags is a list of strings with the following syntax
    <seqID:tag_word:value1-probability,value2-probability>.

    Parameters:
    -----------
    graph: nx.Graph
    tags: list[str]
        list of tags
    seed: float
        random seed to be used in the selection
    """
    for tag in tags:
        seq_ID, attr, attr_probs = tag.split(':')
        attribute_values = []
        weights = []

        for attr_prob in attr_probs.split(','):
            value, prob = attr_prob.split("-")
            attribute_values.append(value)
            weights.append(float(prob))

        idx_nodes = find_atoms(graph, "seqid", int(seq_ID))
        _random_replace_nodes_attribute(graph, attribute_values,
                                        weights, attr,
                                        nodes=idx_nodes, seed=seed)

def gen_seq(args):
    """
    Given a sequence definition and macros defining smaller
    building blocks of the sequence, create a residue graph
    and write it to a .json file to be used with gen_itp.
    """
    macros = {}

    if args.from_file:
        force_field = load_library("seq", None, args.inpath)
        for tag_name in args.from_file:
            tag, name = tag_name.split(":")
            macros[tag] = MacroFile(name, force_field)

    for macro_string in args.macros:
        macro = MacroString(macro_string)
        macros[macro.name] = macro

    seq_graph = generate_seq_graph(args.seq, macros, args.connects)

    _apply_termini_modifications(seq_graph, args.modifications)
    _tag_nodes(seq_graph, args.tags)

    g_json = json_graph.node_link_data(seq_graph)
    with open(args.outpath, "w") as file_handle:
        json.dump(g_json, file_handle, indent=2)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides a class used to describe a gromacs topology and all assciated data.
"""
import os
from pathlib import Path
from collections import defaultdict
from itertools import combinations
import numpy as np
from vermouth.system import System
from vermouth.forcefield import ForceField
from vermouth.gmx.gro import read_gro
from vermouth.pdb import read_pdb
from .top_parser import read_topology
from .linalg_functions import center_of_geometry

COORD_PARSERS = {"pdb": read_pdb,
                 "gro": read_gro}

# small wrapper that is neccessiataed
# by the fact that gro and pdb readers
# return a molecule and a list respectively


def _coord_parser(path, extension):
    reader = COORD_PARSERS[extension]
    molecules = reader(path, exclude=())
    if extension == "pdb":
        molecule = molecules[0]
        for new_mol in molecules[1:]:
            molecule.merge_molecule(new_mol)
    else:
        molecule = molecules

    return molecule


def replace_defined_interaction(interaction, defines):
    """
    Given an `interaction` check if parameters
    are defined in a list of defines and replace
    by the corresponding numeric value.

    Parameters
    -----------
    interaction: :tuple:`vermouth.molecule.Interaction`
    defines:  dict
      dictionary of type [define]:value

    Returns
    --------
    interaction
      interaction with replaced defines
    """
    new_parameters = []
    for parameter in interaction.parameters:
        if parameter in defines:
            values = defines[parameter]
            for new_param in values:
                new_parameters.append(new_param)
        else:
            new_parameters.append(parameter)

    interaction.parameters[:] = new_parameters[:]

    return interaction


def lorentz_berthelot_rule(sig_A, sig_B, eps_A, eps_B):
    """
    Lorentz-Berthelot rules for combining LJ paramters.

    Parameters
    -----------
    sig_A:  float
    sig_B:  float
        input sigma values
    eps_A:  float
    eps_B:  float
        input epsilon values

    Returns
    --------
    float
        sigma
    float
        epsilon
    """
    sig = (sig_A + sig_B)/2.0
    eps = (eps_A * eps_B)**0.5
    return sig, eps


def geometric_rule(C6_A, C6_B, C12_A, C12_B):
    """
    Geometric combination rule for combining
    LJ parameters.

    Parameters:
    -----------
    C6_A:  float
    C6_B:  float
        input C6 values
    C12_A:  float
    C12_B:  float
        input C12 values

    Returns:
    --------
    float
         C6
    float
         C12
    """
    C6 = (C6_A * C6_B)**0.5
    C12 = (C12_A * C12_B)**0.5
    return C6, C12


class Topology(System):
    """
    Ties together vermouth molecule definitions, and
    Gromacs topology information.

    Parameters
    ----------
    force_field: :class:`vermouth.forcefield.ForceField`
        A force field object.
    name: str, optional
        The name of the topology.

    Attributes
    ----------
    molecules: list[:class:`~vermouth.molecule.Molecule`]
        The molecules in the system.
    force_field: a :class:`vermouth.forcefield.ForceField`
    nonbond_params: dict
        A dictionary of all nonbonded parameters
    types: dict
        A dictionary of all typed parameter
    defines: list
        A list of everything that is defined
    """

    def __init__(self, force_field, name=None):
        super().__init__(force_field)
        self.name = name
        self.defaults = {}
        self.defines = {}
        self.description = []
        self.atom_types = {}
        self.types = defaultdict(dict)
        self.nonbond_params = {}
        self.mol_idx_by_name = defaultdict(list)

    def preprocess(self):
        """
        Apply all defaults, generate pairs, convert non-bonded
        units. It performs most of the conversion which otherwise
        is done by grompp.
        """
        self.gen_pairs()
        # we need to replace defines before doing bonded interactions
        self.replace_defines()
        self.gen_bonded_interactions()
        # only convert if we not already have sig-eps form
        if self.defaults['comb-rule'] == 1:
            self.convert_nonbond_to_sig_eps()

    def replace_defines(self):
        """
        Replace all interaction paramers with defined parameters.
        """
        # Note that a topology cannot define and generate links so
        # they don't need to be replaced or handled elsewhere
        for block in self.force_field.blocks.values():
            for interactions in block.interactions.values():
                for interaction in interactions:
                    new_interaction = replace_defined_interaction(interaction, self.defines)

    def gen_pairs(self):
        """
        If pairs default is set to yes the non-bond params are
        generated for all pairs according to the combination
        rules. Regardless of if pairs is set or not the self
        interactions from atomtypes are added to `self.nonbond_params`.
        Note that nonbond_params takes precedence over atomtypes and
        generated pairs.
        """
        comb_funcs = {1.0: lorentz_berthelot_rule,
                      2.0: geometric_rule,
                      3.0: lorentz_berthelot_rule}

        comb_rule = comb_funcs[self.defaults["comb-rule"]]

        if self.defaults["gen-pairs"] == "yes":
            for atom_type_A, atom_type_B in combinations(self.atom_types, r=2):
                if frozenset([atom_type_A, atom_type_B]) not in self.nonbond_params:
                    nb1_A = self.atom_types[atom_type_A]["nb1"]
                    nb2_A = self.atom_types[atom_type_A]["nb2"]
                    nb1_B = self.atom_types[atom_type_B]["nb1"]
                    nb2_B = self.atom_types[atom_type_B]["nb2"]
                    nb1, nb2 = comb_rule(nb1_A, nb1_B, nb2_A, nb2_B)
                    self.nonbond_params.update({frozenset([atom_type_A, atom_type_B]):
                                                {"nb1": nb1, "nb2": nb2}})

        for atom_type in self.atom_types:
            if frozenset([atom_type, atom_type]) not in self.nonbond_params:
                nb1 = self.atom_types[atom_type]["nb1"]
                nb2 = self.atom_types[atom_type]["nb2"]
                self.nonbond_params.update({frozenset([atom_type, atom_type]):
                                            {"nb1": nb1, "nb2": nb2}})

    def gen_bonded_interactions(self):
        """
        Check for each interaction if there is
        no parameter for an interaction if that
        parameter is defined in the bonded directive
        of the topology.
        """
        for block in self.force_field.blocks.values():
            for inter_type, interactions in block.interactions.items():
                if inter_type in ["pairs", "exclusions", "virtual_sitesn",
                                  "virtual_sites2", "virtual_sites3", "virtual_sites4"]:
                    continue
                for interaction in interactions:
                    if len(interaction.parameters) == 1:

                        # Some force-fields - in GMX library only OPLS - use bond-type
                        # definitions. Each atomtype matches one bond-type, which
                        # in turn matches an expression in the bondedtypes section
                        if "_FF_OPLS" in self.defines or "_FF_OPLS_AA" in self.defines:
                            atoms = tuple(self.atom_types[block.nodes[node]["atype"]]["bond_type"]
                                         for node in interaction.atoms)
                        # Other force-fields like charmm and amber use the atomtype directly for
                        # matching the bondded types
                        else:
                            atoms = tuple(block.nodes[node]["atype"] for node in interaction.atoms)

                        if atoms in self.types[inter_type]:
                            new_params, meta = self.types[inter_type][atoms]
                        elif atoms[::-1] in self.types[inter_type]:
                            new_params, meta = self.types[inter_type][atoms[::-1]]
                        elif inter_type in "dihedrals" and\
                             ("X", atoms[1], atoms[2], "X") in self.types[inter_type]:
                            new_params, meta = self.types[inter_type][("X", atoms[1], atoms[2], "X")]
                        elif inter_type == "dihedrals" and\
                             ("X", atoms[2], atoms[1], "X") in self.types[inter_type]:
                            new_params, meta = self.types[inter_type][("X", atoms[2], atoms[1], "X")]
                        elif inter_type in "dihedrals" and\
                             (atoms[0], "X", "X", atoms[3]) in self.types[inter_type]:
                            new_params, meta = self.types[inter_type][(atoms[0], "X", "X", atoms[3])]
                        elif inter_type == "dihedrals" and\
                             (atoms[3], "X", "X", atoms[0]) in self.types[inter_type]:
                            new_params, meta = self.types[inter_type][(atoms[3], "X", "X", atoms[0])]
                        else:
                            msg=("In section {} interaction of atoms {} has no corresponding bonded"
                                 "type.")
                            atoms = " ".join(list(map(lambda x: str(x), interaction.atoms)))
                            raise OSError(msg.format(inter_type, atoms))

                        interaction.parameters[:] = new_params[:]
                        if meta:
                            interaction.meta.update(meta)

    def convert_nonbond_to_sig_eps(self):
        """
        Convert all nonbond_params to sigma epsilon form of the
        LJ potential. Note that this assumes the parameters
        are in A, B form.
        """
        for atom_pair in self.nonbond_params:
            nb1 = self.nonbond_params[atom_pair]["nb1"]
            nb2 = self.nonbond_params[atom_pair]["nb2"]

            if nb2 != 0:
                sig = (nb2/nb1)**(1.0/6.0)
            else:
                sig = 0

            if nb1 != 0:
                eps = nb1**2.0/(4*nb2)
            else:
                eps = 0

            self.nonbond_params.update({atom_pair: {"nb1": sig, "nb2": eps}})

    def add_positions_from_file(self, path, build_res=[]):
        """
        Add positions to topology from coordinate file.
        """
        path = Path(path)
        extension = path.suffix.casefold()[1:]
        molecules = _coord_parser(path, extension)
        total = 0
        for meta_mol in self.molecules:
            no_coords = []
            for node in meta_mol.molecule.nodes:
                resname = meta_mol.molecule.nodes[node]["resname"]
                if resname in build_res:
                   no_coords.append(node)
                   continue
                try:
                    position = molecules.nodes[total]["position"]
                except KeyError:
                    no_coords.append(node)
                else:
                    meta_mol.molecule.nodes[node]["position"] = position
                    total += 1

            for node in meta_mol:
                resid = meta_mol.nodes[node]["resid"]
                atoms_in_res = list(meta_mol.nodes[node]["graph"].nodes)

                if all(atom not in no_coords for atom in atoms_in_res):
                    positions = np.array([meta_mol.molecule.nodes[atom]["position"] for
                                          atom in atoms_in_res])
                    center = center_of_geometry(positions)
                    meta_mol.nodes[node]["position"] = center
                    meta_mol.nodes[node]["build"] = False
                else:
                    meta_mol.nodes[node]["build"] = True

    def convert_to_vermouth_system(self):
        system = System()
        system.molecules = []
        system.force_field = self.force_field

        for meta_mol in self.molecules:
            system.molecules.append(meta_mol.molecule)

        return system

    @classmethod
    def from_gmx_topfile(cls, path, name):
        """
        Read a gromacs topology file and return an topology object.

        Parameters
        ----------
        path:  str
           The name of the topology file
        name:  str
           The name of the system
        """
        with open(path, 'r') as _file:
            lines = _file.readlines()

        cwdir = os.path.dirname(path)
        force_field = ForceField(name)
        topology = cls(force_field=force_field, name=name)
        read_topology(lines=lines, topology=topology, cwdir=cwdir)
        return topology



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that build files are properly read.
"""
import textwrap
import pytest
import numpy as np
import vermouth.forcefield
import vermouth.molecule
import polyply
from polyply.src.meta_molecule import Monomer, MetaMolecule
from polyply.src.topology import Topology

@pytest.mark.parametrize('tokens, _type, expected', (
   # for cylinder
   (["PEO", "63", "154", "in", "1", "2", "3", "6", "7"],
    "cylinder",
    {"resname": "PEO", "start": 63, "stop": 154, "parameters":["in", np.array([1.0, 2.0, 3.0]), 6.0, 7.0, "cylinder"]})
   # for recangle
   ,(["PEO", "0", "10", "out", "11", "12", "14", "1", "2", "3"],
    "rectangle",
    {"resname": "PEO", "start": 0, "stop": 10, "parameters":["out", np.array([11.0, 12.0, 13.0]), 1.0, 2.0, 3.0, "rectangle"]})
   # for sphere
   ,(["PEO", "0", "10", "in", "11", "12", "13", "5"],
    "sphere",
    {"resname": "PEO", "start": 0, "stop": 10, "parameters":["in", np.array([11.0, 12.0, 13.0]), 5.0, "sphere"]})
   ))
def test_base_parser_geometry(tokens, _type, expected):
    result = polyply.src.build_file_parser.BuildDirector._base_parser_geometry(tokens, _type)
    assert result.keys() == expected.keys()
    for key in result:
        if key != "parameters":
            assert result[key] == expected[key]
        else:
            assert result[key][0] == expected[key][0]
            assert all(result[key][1] == expected[key][1])
            for result_param, expected_param in zip(result[key][2:], expected[key][2:]):
                assert result_param == expected_param

@pytest.mark.parametrize('line, expected', (
   # simple example
   ("PEO 63 154 5.0 3.0 2.0 50.0",
    {"resname": "PEO", "start": 63, "stop": 154, "parameters": [np.array([5.0, 3.0, 2.0]), 50.]})
   # other example
   ,("PEO 0 10 4.0 4.0 1.0 25.0",
    {"resname": "PEO", "start": 0, "stop": 10, "parameters": [np.array([4.0, 4.0, 1.0]), 25.0]})
   ))
def test_base_parser_geometry(line, expected):
    processor = polyply.src.build_file_parser.BuildDirector([])
    processor.current_molidxs = [1]
    processor.current_molname = "PEO"
    processor._rw_restriction(line)
    result =  processor.rw_options[("PEO", 1)]
    assert result.keys() == expected.keys()
    for key in processor.rw_options[("PEO", 1)]:
        if key != "parameters":
            assert result[key] == expected[key]
        else:
            assert all(result[key][0] == expected[key][0])
            assert result[key][1] == expected[key][1]

@pytest.fixture
def test_molecule():
    # dummy vermouth force-field
    force_field = vermouth.forcefield.ForceField(name='test_ff')
    # monomers used in the meta-molecule
    ALA = Monomer(resname="ALA", n_blocks=4)
    GLU = Monomer(resname="GLU", n_blocks=2)
    THR = Monomer(resname="THR", n_blocks=3)
    # two meta-molecules
    mol = MetaMolecule.from_monomer_seq_linear(force_field,
                                               [ALA, GLU, THR],
                                               "AA")
    return mol

@pytest.mark.parametrize('_type, option, expected', (
   # tag all ALA nodes
   ("cylinder",
    {"resname": "ALA", "start": 1, "stop": 4, "parameters":["in", np.array([5.0, 5.0, 5.0]), 5.0, 5.0]},
    [0, 1, 2, 3]),
   # tag all nodes 1-4 that have name ALA, GLU, THR should not get tagged
   ("cylinder",
    {"resname": "ALA", "start": 1, "stop": 9, "parameters":["in", np.array([5.0, 5.0, 5.0]), 5.0, 5.0]},
   [0, 1, 2, 3]),
   # tag only GLU nodes
   ("sphere",
    {"resname": "GLU", "start": 5, "stop": 6, "parameters":["in", np.array([10.0, 10.0, 10.0]), 5.0]},
   [4, 5]),
   # tag no nodes based on the resid rensame correspondance
   ("sphere",
    {"resname": "GLU", "start": 1, "stop": 4, "parameters":["in", np.array([10.0, 10.0, 10.0]), 5.0]},
   [])
   # tag no nodes based on no resname in molecule
   ,("sphere",
    {"resname": "PPI", "start": 1, "stop": 4, "parameters":["in", np.array([10.0, 10.0, 10.0]), 5.0]},
   [])
   ))
def test_tag_nodes(test_molecule, _type, option, expected):
    polyply.src.build_file_parser.BuildDirector._tag_nodes(test_molecule, _type, option)
    for node in test_molecule.nodes:
        if "restraints" in test_molecule.nodes[node]:
           assert node in expected

@pytest.fixture()
def test_system():
  """
  Create a dummy test system with three types of molecules AA, BB and
  NA. NA is the molecule to be used a ligand. AA and BB are composed
  of different residues.
  """
  # dummy vermouth force-field
  force_field = vermouth.forcefield.ForceField(name='test_ff')
  # monomers used in the meta-molecule
  ALA = Monomer(resname="ALA", n_blocks=4)
  GLU = Monomer(resname="GLU", n_blocks=2)
  THR = Monomer(resname="THR", n_blocks=3)
  # two meta-molecules
  meta_mol_A = MetaMolecule.from_monomer_seq_linear(force_field,
                                                    [ALA, GLU, THR],
                                                    "AA")
  meta_mol_B = MetaMolecule.from_monomer_seq_linear(force_field,
                                                    [GLU, ALA, THR],
                                                    "BB")
  NA = MetaMolecule()
  NA.add_monomer(current=0, resname="NA", connections=[])
  molecules = [meta_mol_A, meta_mol_A.copy(),
               meta_mol_B.copy(), NA, NA.copy(),
               NA.copy(), NA.copy()]
  top = Topology(force_field=force_field)
  top.molecules = molecules
  top.mol_idx_by_name = {"AA":[0, 1], "BB": [2], "NA":[3, 4, 5, 6]}
  return top

@staticmethod
@pytest.mark.parametrize('lines, tagged_mols, tagged_nodes', (
   # basic test
   ("""
   [ molecule ]
   ; name from to
   AA    0  2
   ;
   [ cylinder ]
   ; resname start stop  inside-out  x  y  z  r   z
   ALA   2    4  in  5  5  5  5  5
   """,
   [0, 1],
   [0, 1, 2, 3]),
   # test that a different molecule is tagged
   ("""
   [ molecule ]
   BB  2  3
   [ rectangle ]
   ; resname start stop  inside-out  x  y  z a b c
   ALA   3    6  in  5  5  5  5  5 5
   """,
   [2],
   [2, 3, 4, 5]),
   # test nothing is tagged based on the molname
   ("""
   [ molecule ]
   CC 1 6
   [ sphere ]
   ; resname start stop  inside-out  x  y  z r
   ALA   2    4  in  5  5  5  5
   """,
   [],
   []),
   ))
def test_parser(test_system, lines, tagged_mols, tagged_nodes):
   lines = textwrap.dedent(lines).splitlines()
   ff = vermouth.forcefield.ForceField(name='test_ff')
   polyply.src.build_file_parser.read_build_file(lines,
                                                 test_system.molecules)
   for idx, mol in enumerate(test_system.molecules):
       for node in mol.nodes:
           if "restraints" in mol.nodes[node]:
               assert node in tagged_nodes
               assert idx in tagged_mols




# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import defaultdict
from itertools import combinations
import networkx as nx
from tqdm import tqdm
import vermouth.molecule
from vermouth.log_helpers import StyleAdapter, get_logger
from vermouth.molecule import Interaction, attributes_match
from vermouth.graph_utils import make_residue_graph
from vermouth.processors.do_links import match_order
from .processor import Processor
from .graph_utils import neighborhood

LOGGER = StyleAdapter(get_logger(__name__))

class MatchError(Exception):
    """Raised we find no match between links and molecule"""

def expand_excl(molecule):
    """
    Given a `molecule` add exclusions for nodes that
    have the exclude attribute.

    Parameters:
    -----------
    molecule: :class:`vermouth.molecule`

    Returns:
    --------
    :class:`vermouth.molecule`
    """
    exclude = nx.get_node_attributes(molecule, "exclude")
    nrexcl = molecule.nrexcl
    had_excl=[]
    for node, excl in exclude.items():
        if excl > nrexcl:
            excluded_nodes = neighborhood(molecule, node, max_length=excl, min_length=nrexcl)
            for ndx in excluded_nodes:
                excl = Interaction(atoms=[node, ndx],
                                   parameters=[],
                                   meta={})
                if frozenset([node, ndx]) not in had_excl and node != ndx:
                    had_excl.append(frozenset([node, ndx]))
                    molecule.interactions["exclusions"].append(excl)
    return molecule

def find_atoms(molecule, ignore=[], **attrs):
    """
    Yields all indices of atoms that match `attrs`

    Parameters
    ----------
    molecule: :class:`vermouth.molecule.Molecule`
    **attrs: collections.abc.Mapping
        The attributes and their desired values.

    Yields
    ------
    collections.abc.Hashable
        All atom indices that match the specified `attrs`
    """
    for node_idx in molecule:
        node = molecule.nodes[node_idx]
        if vermouth.molecule.attributes_match(node, attrs, ignore_keys=ignore):
            yield node_idx

def _build_link_interaction_from(molecule, interaction, match):
    """
    Creates a new interaction from a link type interaction
    and a match specification for a molecule. Note that
    it can overwrite equivalent interactions.

    Parameters
    ----------
    molecule: :class:`vermouth.molecule.Molecule`
    interaction: :type:`vermoth.molecule.Interaction`
    match: :type:`dict`
        The mapping between the interaction atom names
        and the molecule atom names.

    Returns
    ------
    `vermouth.molecule.Interaction`
    """
    atoms = tuple(match[idx] for idx in interaction.atoms)
    parameters = [
        param(molecule, match) if callable(param) else param
        for param in interaction.parameters
    ]
    new_interaction = interaction._replace(
        atoms=atoms,
        parameters=parameters
    )
    return new_interaction


def apply_explicit_link(molecule, link):
    """
    Applies interactions from a link regardless of any
    checks. This requires atom keys in the link to be of
    int type. Within polyply the explicit flag can
    be set to these links for the program to know
    to apply them using this method.

    Parameters
    ----------
    molecule: :class:`vermouth.molecule.Molecule`
        A vermouth molecule definintion
    link: :class:`vermouth.molecule.Link`
        A vermouth link definintion
    """
    for inter_type, inter_list in link.interactions.items():
        for interaction in inter_list:
            try:
                interaction.atoms[:] = [int(atom) - 1 for atom in interaction.atoms]
            except ValueError as err:
                msg = """Trying to apply an explicit link but interaction
                      {} but cannot convert the atoms to integers. Note
                      explicit links need to be defined by atom numbers."""
                raise ValueError(msg.format(interaction)) from err
            if set(interaction.atoms).issubset(set(molecule.nodes)):
                molecule.add_or_replace_interaction(inter_type, interaction.atoms,
                                                    interaction.parameters,
                                                    meta=interaction.meta)
                atoms = interaction.atoms
                new_edges = list(zip(atoms[:-1], atoms[1:]))
                molecule.add_edges_from(new_edges)
            else:
                raise IOError("Atoms of link interaction {} are not "
                              "part of the molecule.".format(interaction))

def _check_relative_order(resids, orders):
    """
    This function checks if the relative order of
    list of lists of residues adheres to the order
    specifications of vermouth orders.
    """

    order_match = {}
    for order, resid in zip(orders, resids):
        if order not in order_match:
            order_match[order] = resid
        # Assert all orders correspond to the same resid
        elif order_match[order] != resid:
            return False

    for ((order1, resid1), (order2, resid2)) in combinations(order_match.items(), 2):
        # Assert the differences between resids correspond to what
        # the orders require.
        if not match_order(order1, resid1, order2, resid2):
            return False

    return True

def _assign_link_resids(res_link, match):
    """
    Given a link at residue level (`res_link`) and
    a dict (`match`) specifying to which resids
    the res_link nodes map, create a correspondence
    dict that maps the higher resolution nodes to
    the resids specified in` match`. Each `res_link`
    node by definition can only map to one resid provided
    in `match`. The lower resolution nodes associated
    to a particular res_link node are stored in the 'graph'
    attribute of res_link.
    Note that the nodes in that graph are consecutive
    and uniquely specify to a single node in the graph
    specifying all residues at higher resolution.

    Parameters:
    -----------
    res_link: :class:`nx.Graph`
        must have a 'graph' attribute for each node
        that itself is a graph of the atoms represented
    match: dict
        dict matching a resid to the node_key of res_link

    Returns:
    --------
    :type:dict
        correspondence of high resolution nodes to resid
    """
    link_node_to_resid = {}
    for resid, link_node in match.items():
        for node in res_link.nodes[link_node]['graph']:
            link_node_to_resid[node] = resid
    return link_node_to_resid

def match_link_and_residue_atoms(meta_molecule, link, link_to_resid):
    """
    Given a meta_molecule a link and a correspondence of the link
    nodes and those in the meta_molecule, establish a correspondence
    between the link nodes and those in the higher resolution
    molecule of the meta_molecule. Note that the meta_molecule needs
    to have the "block" attribute describing which atoms a single
    meta_molecule node described.

    Parameters:
    -----------
    meta_molecule: :class:`polyply.src.meta_molecule.MetaMolecule`
    link:          :class:`vermouth.molecule.Link`
    link_to_resid:  :type:dict
        correspondence dict of link nodes to meta_molecule nodes

    Returns:
    --------
    :type:dict
        correspondence dict of link nodes to atoms in the
        the meta_molecule.molecule attribute
    """
    link_to_mol = {}
    for node in link.nodes:
        meta_mol_key = link_to_resid[node]
        block = meta_molecule.nodes[meta_mol_key]["graph"]
        resid = block.nodes[list(block.nodes)[0]]["resid"]
        attrs = link.nodes[node]
        # relative resid has been asserted before so we can
        # exclude it here
        ignore = ['order', 'charge_group', 'replace', 'resid']
        matchs = list(find_atoms(block, ignore=ignore, **attrs))

        if len(matchs) == 1:
            link_to_mol[node] = matchs[0]
        elif len(matchs) == 0:
            msg = "Found no matchs for node {} in resiue {}. Cannot apply link."
            raise MatchError(msg.format(node, resid))
        else:
            msg = "Found {} matches for node {} in resiue {}. Cannot apply link."
            raise MatchError(msg.format(len(matchs), node, resid))

    return link_to_mol

def _res_match(node1, node2):
    """
    Helper function which returns true if the resname
    attribute of two nodes matches and false otherwise.
    This function correctly handles choice objects
    for the resname.

    Parameters:
    -----------
    node1:  :type:dict
        attribute dict of node
    node2:  :type:dict
        attribute dict of node

    Returns:
    --------
    bool
    """
    # this is not equivalent to comparing just the resname
    # attributes because resname can be a choice object
    # which at the moment does not compare properly
    ignore = [key for key in node2.keys() if key != "resname"]
    return attributes_match(node1, node2, ignore_keys=ignore)

def _linktype_match(edge_attrs1, edge_attrs2):
    type1 = edge_attrs1.get("linktype", None)
    type2 = edge_attrs2.get("linktype", None)
    return type1 == type2

def _resnames_match(resnames, allowed_resnames):
    """
    Return true if one element in resnames matches
    one element in allowed_resnames.

    Parameters
    ----------
    resnames: `abc.iterable`
    allowed_resnames: `abc.iterable`
    """
    for resname in resnames:
        if resname in allowed_resnames:
           return True
    return False

def _get_link_resnames(link):
    """
    Get the resnames of a `link` directly from the node attributes
    of the link atoms. It is not safe enough to just take the
    link name, because it is often undefined or empty. Note that
    a link name can also be a `vermouth.molecule.Choice` object.
    Parameters
    ----------
    link: :class:`vermouth.molecule.Link`
    Returns
    ----------
    set
      all unique resnames of a molecule
    """
    res_names = list(nx.get_node_attributes(link, "resname").values())
    out_resnames = set()

    for name in res_names:
        if isinstance(name, vermouth.molecule.LinkPredicate) and not isinstance(name.value, str):
            out_resnames.update(name.value)
        else:
            out_resnames.add(name)

    return out_resnames

class ApplyLinks(Processor):
    """
    This processor takes a a class:`polyply.src.MetaMolecule` and
    based on links defined in the `force-field` attribute of the
    MetaMolecule applies them when appropiate.

    Links are defined as connections between two residues at the
    high resolution level whereas the MetaMolecule is the molecule
    at residue level or higher degree of CG. The algorithm proceeds
    by first coarse-graining a link to residue level. Subsequently
    it is checked based on the connectivity of the link at residue
    level and the residue name to which residues in the MetaMolecule
    this link fits. Subsequently, based on the high resolution
    definition the algorithm looks up the atoms in the matching residues
    corresponding to the link. If all atoms are present the link is
    applied otherwise the link is not applied.

    Note that because links can overwrite each other they are first
    stored in the instance variable `applied_links` and only written
    to the MetaMolecule.molecule graph after all links have been
    checked and overwritten if needed. This means the apply_links_between_residues
    method only populates this instance variable, while write to
    molecule applies those links.
    """
    def __init__(self, *args, debug=False, **kwargs):
        self.debug = debug
        super().__init__(*args, **kwargs)
        self.applied_links = defaultdict(dict)

    def apply_link_between_residues(self, meta_molecule, link, link_to_resid):
        """
        Applies a link between specific residues, if and only if
        the link atoms (incl. all attributes) match at most one atom
        in a respective link. It adds the link to the applied_links
        instance variable, from which later the links are added to
        the molecule. Note that replace statements are already update
        the molecule, as they potentially apply to consecutive links.
        Edges are also updated in place.

        Parameters
        ----------
        meta_molecule: :class:`polyply.src.MetaMolecule`
        link: :class:`vermouth.molecule.Link`
            A vermouth link definition
        link_to_resids: :type:dict
            a dict matching link nodes to a resid in meta_molecule
        """
        # handy variable for later referencing
        molecule = meta_molecule.molecule

        # look-up which atoms in the link correspond
        # to which atoms in the molecule this function
        # raises an MatchError if no atoms are found

        link_to_mol = match_link_and_residue_atoms(meta_molecule,
                                                   link,
                                                   link_to_resid)

        # if all atoms have a match the link applies and we first
        # replace any attributes from the link node section
        for node in link.nodes:
            molecule.nodes[link_to_mol[node]].update(link.nodes[node].get('replace', {}))

        # based on the match we build the link interaction
        for inter_type in link.interactions:
            for interaction in link.interactions[inter_type]:
                new_interaction = _build_link_interaction_from(molecule, interaction, link_to_mol)
                # it is not guaranteed that interaction.atoms is a tuple
                # the key is the atoms involved in the interaction and the version type so
                # that multiple versions are kept and not overwritten
                interaction_key = tuple(new_interaction.atoms) +\
                                  tuple([new_interaction.meta.get("version",1)])
                self.applied_links[inter_type][interaction_key] = (new_interaction, link.citations)
        # now we already add the edges of this link
        # links can overwrite each other but the edges must be the same
        # this is safer than using the make_edge method because it accounts
        # for edges written in the edges directive
        for edge in link.edges:
            molecule.add_edge(link_to_mol[edge[0]], link_to_mol[edge[1]])

    def run_molecule(self, meta_molecule):
        """
        Given a meta_molecule the function iterates over all edges
        (i.e. pairs of residues) and finds all links that fit based
        on residue-name and order attribute. Subsequently it tries
        to apply these links. If a MatchError is encountered the
        link is not applied. The meta_molecule is updated in place
        but also returned.

        Parameters
        ----------
        molecule: :class:`polyply.src.meta_molecule.MetaMolecule`
             The meta molecule to process.

        Returns
        ---------
        :class:`polyply.src.meta_molecule.MetaMolecule`
        """
        molecule = meta_molecule.molecule
        force_field = meta_molecule.force_field
        resnames = set(nx.get_node_attributes(molecule, "resname").values())
        for link in tqdm(force_field.links):
            link_resnames = _get_link_resnames(link)
            if not _resnames_match(resnames, link_resnames):
                continue
            # we only use the order because each order needs to be
            # matching exactly 1 residue, which means their resname
            # needs to match as well. However, resname can be a
            # choice object which is not hashable so we don't
            # use resname in constructing the res-graph.
            res_link = make_residue_graph(link, attrs=('order',))
            # however when finding the LCIS we do match against the residue
            # name and topology of the link
            GM = nx.isomorphism.GraphMatcher(meta_molecule,
                                             res_link,
                                             node_match=_res_match,
                                             edge_match=_linktype_match)
            raw_matchs = GM.subgraph_isomorphisms_iter()
            for match in raw_matchs:
                nodes = match.keys()
                resids =[meta_molecule.nodes[node]["resid"] for node in nodes]
                orders = [ res_link.nodes[match[node]]["order"] for node in nodes]
                if _check_relative_order(resids, orders):
                    link_node_to_resid = _assign_link_resids(res_link, match)
                    try:
                        self.apply_link_between_residues(meta_molecule, link, link_node_to_resid)
                    except MatchError as error:
                        LOGGER.debug(str(error), type='step')

        for inter_type in self.applied_links:
            for interaction, citation in self.applied_links[inter_type].values():
                meta_molecule.molecule.interactions[inter_type].append(interaction)
                if citation:
                    meta_molecule.molecule.citations.update(citation)

        for link in force_field.links:
            if link.molecule_meta.get('by_atom_id'):
                apply_explicit_link(molecule, link)

        expand_excl(molecule)
        return meta_molecule



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test linear algebra aux functions.
"""

import textwrap
import pytest
import numpy as np
import math
import networkx as nx
import polyply
from polyply.src.linalg_functions import (_u_vect, _angle,
                                         _dih, _radius_of_gyration,
                                         center_of_geometry, _vector_angle_degrees)


def test_vector_angle_degrees():
    v1 = np.array([0.0, 0.0, 0.0])
    v2 = np.array([0.0, 0.0, 2.0])
    v3 = np.array([0.0, 2.0, 2.0])
    assert math.isclose(_vector_angle_degrees(v1-v2, v2-v3), 90.0)

def test_u_vect():
    v1 = np.array([0.0, 0.0, 0.0])
    v2 = np.array([0.0, 0.0, 2])
    u = _u_vect(v1-v2)
    assert math.isclose(np.linalg.norm(u), 1.)

def test_angle():
    v1 = np.array([0.0, 0.0, 0.0])
    v2 = np.array([0.0, 0.0, 2.0])
    v3 = np.array([0.0, 2.0, 2.0])
    assert math.isclose(_angle(v1, v2, v3), 90.0)

def test_dih():
    v1 = np.array([0.0, 0.0, 0.0])
    v2 = np.array([0.0, 0.0, 2.0])
    v3 = np.array([0.0, 2.0, 2.0])
    v4 = np.array([0.0,-2.0, 0.0])
    assert math.isclose(_dih(v1, v2, v3, v4), 0.0)

def test_geometrical_center():
   coords = np.array([[0.0, 0.0, 1.0],
                      [0.0, 1.0, 0.0],
                      [1.0, 0.0, 0.0],
                      [-1.0, 0.0, 0.0],
                      [0.0, -1.0, 0.0],
                      [0.0, 0.0, -1.0]])

   center = center_of_geometry(coords)
   assert math.isclose(np.linalg.norm(center), 0.0)

def test_radius_of_gyration():
    coords = np.array([[0.0, 0.0, 1.0],
                       [0.0, 1.0, 0.0],
                       [1.0, 0.0, 0.0],
                       [-1.0, 0.0, 0.0],
                       [0.0, -1.0, 0.0],
                       [0.0, 0.0, -1.0]])

    rg = _radius_of_gyration(coords)
    assert math.isclose(rg, 1.0)

@pytest.mark.parametrize('vectors, expected',(
                        # single column vector with floats
                        ([np.array([[2.5], [2.5], [2.5]])],
                         np.array([[15.0], [37.5], [60.0]])),
                        # simple 2x3 matrix with floats
                        ([np.array([[2.5, 3.5], [2.5, 3.5] , [2.5, 3.5]])],
                          np.array([[15.0, 21.0], [37.5, 52.5], [60.0 , 84.0]])),
                        # multiple column vectors
                        ([np.array([[2.5], [2.5], [2.5]]),
                          np.array([[3.5], [3.5], [3.5]])],
                         np.array([[52.5], [131.25], [210.0]])),
                        # multiple matrices
                        ([np.array([[2.5, 4.5], [2.5, 4.5], [2.5, 4.5]]),
                          np.array([[3.5, 5.5], [3.5, 5.5], [3.5, 5.5]])],
                         np.array([[147.0,  231.0], [367.5, 577.5], [588.0, 924.0]])),
                        ))
def test_matrix_multiplication(vectors, expected):
        matrix = np.array([[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]])
        result = polyply.src.linalg_functions._matrix_multiplication(matrix, *vectors)
        assert result.shape == expected.shape
        assert np.allclose(expected, result)

@pytest.mark.parametrize('vectors, angles, expected',(
                        # single vector rotation around x
                        (np.array([[2.0], [3.5], [4.5]]),
                         np.array([90, 0.0, 0.0]),
                         np.array([[2.0], [-4.5], [3.5]])),
                        # single vector rotation around y
                        (np.array([[2.5], [3.5], [4.5]]),
                         np.array([0.0, 90.0, 0.0]),
                         np.array([[4.5], [3.5], [-2.5]])),
                        # single vector rotation around z
                        (np.array([[2.5], [3.5], [4.5]]),
                         np.array([0.0, 0.0, 90.0]),
                         np.array([[-3.5], [2.5], [4.5]])),
                        # single vector rotation around y, z
                        (np.array([[2.5], [3.5], [4.5]]),
                         np.array([0.0, 90.0, 90.0]),
                         np.array([[-3.5], [4.5], [-2.5]])),
                        # multiple points rotation around x, y
                        (np.array([[2.5, 5.5], [3.5, 6.5], [4.5, 7.5]]),
                         np.array([90.0, 90.0, 0.0]),
                         np.array([[3.5, 6.5], [-4.5, -7.5], [-2.5, -5.5]])),
                        ))
def test_rotation_matrix(vectors, angles, expected):
        result = polyply.src.linalg_functions._rotate_xyz(vectors, *np.deg2rad(angles))
        assert result.shape == expected.shape
        assert np.allclose(expected, result)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import textwrap
import argparse
import pytest
import networkx as nx
import vermouth.forcefield
import vermouth.molecule
import vermouth.gmx.itp_read
from polyply import gen_params, TEST_DATA

class TestGenParams():
    @staticmethod
    @pytest.mark.parametrize('args_in, ref_file', (
        (["-f", TEST_DATA + "/gen_params/input/PEO.martini.3.itp"
         ,"-seq", "PEO:10", "-name", "PEO", "-o",
         TEST_DATA + "/gen_params/output/PEO_out.itp"],
         TEST_DATA + "/gen_params/ref/PEO_10.itp"),
        (["-f", TEST_DATA + "/gen_params/input/PS.martini.2.itp",
         "-seqf", TEST_DATA + "/gen_params/input/PS.json",
         "-name", "PS",
         "-o", TEST_DATA + "/gen_params/output/PS_out.itp"]
         ,
         TEST_DATA + "/gen_params/ref/PS_10.itp"),
        (["-f", TEST_DATA + "/gen_params/input/P3HT.martini.2.itp",
         "-seq", "P3HT:10",
         "-name", "P3HT",
         "-o", TEST_DATA + "/gen_params/output/P3HT_out.itp"],
         TEST_DATA + "/gen_params/ref/P3HT_10.itp"),
        (["-f", TEST_DATA + "/gen_params/input/PPI.ff",
         "-seqf", TEST_DATA + "/gen_params/input/PPI.json",
         "-name", "PPI",
         "-o", TEST_DATA + "/gen_params/output/PPI_out.itp"],
         TEST_DATA + "/gen_params/ref/G3.itp"),
        (["-f", TEST_DATA + "/gen_params/input/test.ff",
         "-seq", "N1:1", "N2:1", "N1:1", "N2:1", "N3:1",
         "-name", "test",
         "-o", TEST_DATA + "/gen_params/output/test_out.itp"],
         TEST_DATA + "/gen_params/ref/test_rev.itp"),
        # check if edge attributes are parsed and properly applied
        (["-f", TEST_DATA+"/gen_params/input/test_edge_attr.ff",
         "-seqf", TEST_DATA + "/gen_params/input/test_edge_attr.json",
         "-name", "test",
         "-o", TEST_DATA + "/gen_params/output/test_edge_attr_out.itp"],
         TEST_DATA + "/gen_params/ref/test_edge_attr_ref.itp")
        ))
    def test_gen_params(args_in, ref_file):
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        parser.add_argument('-name', required=True, type=str, dest="name",
                            help="name of the final molecule")
        file_group = parser.add_argument_group('Input and output files')
        file_group.add_argument('-f', dest='inpath', required=False, type=Path,
                                help='Input file (ITP|FF)', nargs="*")
        file_group.add_argument('-o', dest='outpath', type=Path,
                                help='Output ITP (ITP)')
        file_group.add_argument('-seq', dest='seq', required=False, nargs='+',
                                type=str, help='linear sequence')
        file_group.add_argument('-seqf', dest='seq_file', required=False, type=Path,
                                help='linear sequence')
        ff_group = parser.add_argument_group('Force field selection')
        ff_group.add_argument('-lib', dest='lib', required=False, type=str,
                              help='force-fields to include from library', nargs='*')
        ff_group.add_argument('-ff-dir', dest='extra_ff_dir', action='append',
                              type=Path, default=[],
                              help='Additional repository for custom force fields.')
        ff_group.add_argument('-list-lib', action='store_true', dest='list_ff',
                              help='List all known force fields, and exit.')
        ff_group.add_argument('-list-blocks', action='store_true', dest='list_blocks',
                              help='List all Blocks known to the'
                              ' force field, and exit.')
        ff_group.add_argument('-list-links', action='store_true', dest='list_links',
                              help='List all Links known to the'
                              ' force field, and exit.')

        args = parser.parse_args(args_in)
        gen_params(args)

        force_field = vermouth.forcefield.ForceField(name='test_ff')

        for path_name in [args.outpath, ref_file]:
            with open(path_name, 'r') as _file:
                lines = _file.readlines()
            vermouth.gmx.itp_read.read_itp(lines, force_field)

        ref_name = args.name + "ref"

        #1. Check that all nodes and attributes are the same
        assert set(force_field.blocks[ref_name].nodes) == set(force_field.blocks[args.name].nodes)
        for node in force_field.blocks[ref_name].nodes:
            ref_attrs = nx.get_node_attributes(force_field.blocks[ref_name], node)
            new_attrs = nx.get_node_attributes(force_field.blocks[args.name], node)
            assert new_attrs == ref_attrs

        #2. Check that all interactions are the same
        int_types_ref = force_field.blocks[ref_name].interactions.keys()
        int_types_new = force_field.blocks[args.name].interactions.keys()
        assert int_types_ref == int_types_new

        for key in force_field.blocks[ref_name].interactions:
            print(key)
            for term in force_field.blocks[ref_name].interactions[key]:
                print(term)
                assert term in force_field.blocks[args.name].interactions[key]



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from numpy import sqrt, pi, cos, sin, dot, cross, arccos, degrees
from numpy.linalg import norm
from scipy.spatial.transform import Rotation
from polyply import jit

def _vector_angle_degrees(v1, v2):
    """
    Compute the angle between two vectors
    in degrees and between 0 180 degrees.

    Parameters
    ----------
    v1: np.array
    v2: np.array

    Returns
    ---------
    float
      angle in degrees
    """
    angle = degrees(arccos(dot(u_vect(v1), u_vect(v2))))
    return angle

# this is the numba implementation
vector_angle_degrees = jit(_vector_angle_degrees)

def _u_vect(vect):
    """
    Compute unit vector of vector.

    Parameters
    ----------
    vect: np.array

    Returns
    -----------
    np.array
    """
    u_vect = vect/norm(vect)
    return u_vect

# this is the numba implementation
u_vect = jit(_u_vect)

def _angle(A, B, C):
    """
    Compute angle between three points,
    where `B` is the center of the angle.

    Paramters
    ---------
    A, B, C:  numpy.array

    Returns
    ---------
    float
         angle in degrees
    """
    v1 = B - A
    v2 = B - C
    angle = vector_angle_degrees(v1, v2)
    return angle
angle = jit(_angle)

def _dih(A, B, C, D):
    """
    Compute dihedral angle between four points,
    where `B` and `C` are in the center.

    Paramters
    ---------
    A, B, C, D:  numpy.array

    Returns
    ---------
    float
         angle in degrees
    """
    r1 = A - B
    r2 = B - C
    r3 = C - D
    n1 = u_vect(cross(r1, r2))
    n2 = u_vect(cross(r2, r3))
    dih = vector_angle_degrees(n1, n2)
    return dih
dih = jit(_dih)

def center_of_geometry(points):
    """
    Compute center of geometry.

    Paramters
    ---------
    points:  numpy.array

    Returns
    ---------
    float
    """
    COG = np.average(points, axis=0)
    return COG

def norm_sphere(values=50):
    """
    Generate unit vectors on a
    sphere. Note the unit vectors
    cover the sphere with equal probablility
    around the sphere.

    For more information see:
    https://stackoverflow.com/questions/33976911/generate-a-
    random-sample-of-points-distributed-on-the-surface-of-a-unit-sphere

    https://mathworld.wolfram.com/SpherePointPicking.html

    Paramters
    ---------
    values: int
        number of vectors to generate

    Returns
    ---------
    np.array shape (values, 3)
    """
    v_sphere = np.random.normal(0.0, 1, (values, 3))
    return np.array([u_vect(vect) for vect in v_sphere])

def _radius_of_gyration(points):
    """
    Compute radius of gyration of points.

    Parameters
    ---------
    points: np.array

    Returns
    ---------
    float
         radius of gyration
    """
    N = len(points)
    diff=np.zeros((N**2))
    count=0
    for i in points:
        for j in points:
            diff[count]=np.dot((i - j),(i-j))
            count = count + 1
    RG = np.sqrt(1/(2*N**2.0) * np.sum(diff))
    return RG
radius_of_gyration = jit(_radius_of_gyration)

def _matrix_multiplication(*args):
   """
   Implements matrix multiplicaiton of numpy ndarrays.
   Basically the same as np.matmul but it can handle 1-D
   column vectors. In addition together with numba a
   modest speed-up is expected. See the following for
   some testing https://stackoverflow.com/questions/36526708/\
   comparing-python-numpy-numba-and-c-for-matrix-multiplication
   Also note that the data type of the matrices needs to be
   explicitly set and is assumed to be np.float64.

   Parameters
   ----------
   args: list[np.array]
       a list of numpy ndarrays
   """
   matrix_a = args[0]
   for matrix_b in args[1:]:
       rows, other_dim = matrix_a.shape
       columns = matrix_b.shape[1]
       new_matrix = np.zeros((rows, columns))
       for i in range(0, rows):
           for j in range(0, columns):
               for k in range(0, other_dim):
                   new_matrix[i,j] += matrix_a[i,k] * matrix_b[k,j]

       matrix_a = new_matrix
   return new_matrix

matrix_multiplication = jit(_matrix_multiplication)

def _rotate_xyz(object_xyz, theta_x, theta_y, theta_z):
    """
    Rotate `object_xyz` around angles `theta_x`, `theta_y`
    and `theta_z` around the x,y,z axis respectively and
    return coordinates of the rotated object. Note
    object_xyz needs to be in column vector format i.e.
    of numpy shape (3, N).

    Parameters:
    ----------
    object_xyz: numpy.ndarray
        coordinates of the object
    thata_x: float
    thata_y: float
    thata_z: float
        angles in degrees
    """
    # convert all angles to degree
    #theta_z, theta_y, theta_x = np.deg2rad(theta_z), np.deg2rad(theta_y), np.deg2rad(theta_x)
    # compute rotatin matrix around z
    sin_z = np.sin(theta_z)
    cos_z = np.cos(theta_z)
    rot_z = np.array([[cos_z, -1.0*sin_z, 0.0], [sin_z, cos_z , 0.0], [0.0, 0.0, 1.0]], dtype=np.float64)
    # compute rotation matrix around y
    cos_y = np.cos(theta_y)
    sin_y = np.sin(theta_y)
    rot_y = np.array([[cos_y, 0.0, sin_y], [0.0, 1.0, 0.0], [-sin_y, 0.0, cos_y]], dtype=np.float64)
    # compute rotation matrix around x
    cos_x = np.cos(theta_x)
    sin_x = np.sin(theta_x)
    rot_x = np.array([[1.0, 0.0, 0.0], [0.0, cos_x, -sin_x], [0.0, sin_x, cos_x]], dtype=np.float64)
    # consecutively multiply all matrices and the object matrix
    rotated_object = matrix_multiplication(rot_z, rot_y, rot_x, object_xyz)
    return rotated_object

rotate_xyz = jit(_rotate_xyz)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that ligands are properly annotated.
"""
import pytest
import numpy as np
import networkx as nx
import vermouth
import polyply
from polyply.src.meta_molecule import Monomer, MetaMolecule
from polyply.src.topology import Topology

# <mol_name>#<mol_idx>-<resname>#<resid>
@pytest.mark.parametrize('spec, expected', [
    ('', {}),
    ('A-ALA#1', {'molname': 'A', 'resname': 'ALA', 'resid': 1}),
    ('A-ALA1', {'molname': 'A', 'resname': 'ALA1'}),
    ('-ALA1', {'resname': 'ALA1'}),
    ('-ALA#1', {'resname': 'ALA', 'resid':1}),
    ('ALA1', {'molname': 'ALA1'}),
    ('A-ALA', {'molname': 'A', 'resname': 'ALA'}),
    ('ALA', {'molname': 'ALA'}),
    ('ALA#5', {'molname': 'ALA', 'mol_idx':5}),
    ('A#1-ALA#1', {'molname': 'A', 'mol_idx':1, 'resname': 'ALA', 'resid': 1.0}),
    ('#1-ALA#1', {'mol_idx':1, 'resname': 'ALA', 'resid': 1.0}),
    ('#2', {'mol_idx': 2}),
    ('#2-ALA', {'mol_idx': 2, 'resname': 'ALA'}),
    ('#2-#1', {'mol_idx': 2, 'resid':1.0}),
])
def test_parse_residue_spec(spec, expected):
    found = polyply.src.annotate_ligands.parse_residue_spec(spec)
    assert found == expected

# <mol_name>#<mol_idx>-<resname>#<resid>
@pytest.mark.parametrize('spec', [
    '#-#',
    '#-#2',
    'A#-#',
    '-A#qds',
     ])
def test_parse_residue_spec_fail(spec):
    with pytest.raises(IOError):
         polyply.src.annotate_ligands.parse_residue_spec(spec)

@pytest.fixture
def example_molecule():
    mol = nx.Graph()
    nodes = [
        {'resname': 'ALA', 'resid': 1, 'atomname':'BB'},
        {'resname': 'ALA', 'resid': 1, 'atomname':'SC1'},
        {'resname': 'ALA', 'resid': 2, 'atomname':'BB'},
        {'resname': 'ALA', 'resid': 2, 'atomname':'SC1'},
        {'resname': 'GLY', 'resid': 3, 'atomname':'BB'},
        {'resname': 'GLY', 'resid': 3, 'atomname':'SC1'},
    ]
    mol.add_nodes_from(enumerate(nodes))
    mol.add_edges_from([(0, 1), (0, 2), (2, 3), (2, 4), (4, 5)])
    return mol

@pytest.mark.parametrize('mol_attrs, expected', (
    ({"resid": 1}, [0, 1]),
    ({"resid": 2}, [2, 3]),
    ({"resname": "ALA"}, [0, 1, 2, 3]),
    ({"resname": "GLY"}, [4, 5]),
    ({"resid": 1, "resname": "ALA"}, [0, 1]),
    ({"resid": 1, "resname": "GLU"}, []),
    ({}, [0, 1, 2, 3, 4, 5])))
def test_find_nodes(example_molecule, mol_attrs, expected):
    found = list(polyply.src.annotate_ligands._find_nodes(example_molecule, mol_attrs))
    assert found == expected

@pytest.fixture()
def example_system():
    """
    Create a dummy test system with three types of molecules AA, BB and
    NA. NA is the molecule to be used a ligand. AA and BB are composed
    of different residues.
    """
    # dummy vermouth force-field
    force_field = vermouth.forcefield.ForceField(name='test_ff')
    # monomers used in the meta-molecule
    ALA = Monomer(resname="ALA", n_blocks=2)
    GLU = Monomer(resname="GLU", n_blocks=1)
    THR = Monomer(resname="THR", n_blocks=1)
    # two meta-molecules
    meta_mol_A = MetaMolecule.from_monomer_seq_linear(force_field,
                                                      [ALA, GLU, THR],
                                                      "AA")
    meta_mol_B = MetaMolecule.from_monomer_seq_linear(force_field,
                                                      [GLU, ALA, THR],
                                                      "BB")
    NA = MetaMolecule()
    NA.add_monomer(current=0, resname="NA", connections=[])
    molecules = [meta_mol_A, meta_mol_A.copy(),
                 meta_mol_B.copy(), NA, NA.copy(),
                 NA.copy(), NA.copy()]
    top = Topology(force_field=force_field)
    top.molecules = molecules
    top.mol_idx_by_name = {"AA":[0, 1], "BB": [2], "NA":[3, 4, 5, 6]}
    return top

@pytest.mark.parametrize('lig_spec, expected_lig_defs', [
    (('AA-ALA#1', 'NA'), {0: [(0,
                               3,
                             {'molname': 'AA', 'resname': 'ALA', 'resid': 1.0},
                             {'molname': 'NA'})],
                          1: [(0,
                               4,
                             {'molname': 'AA', 'resname': 'ALA', 'resid': 1.0},
                             {'molname': 'NA'})]}),
    (('#0-ALA#1', 'NA'), {0: [(0,
                               3,
                             {'mol_idx': 0.0, 'resname': 'ALA', 'resid': 1.0},
                             {'molname': 'NA'})]}),
   (('#0-ALA#1', '#5-NA'), {0: [(0,
                               5,
                             {'mol_idx': 0.0, 'resname': 'ALA', 'resid': 1.0},
                             {'molname': 'NA'})]}),
    (('AA#1-ALA#1', 'NA'), {1: [(0,
                                 3,
                             {'mol_idx': 1, 'molname': 'AA', 'resname': 'ALA', 'resid': 1.0},
                             {'molname': 'NA'})]}),
    (('AA-GLU', 'NA'), {0: [(2,
                             3,
                            {'molname': 'AA', 'resname': 'GLU'},
                            {'molname': 'NA'})],
                        1: [(2,
                             4,
                            {'molname': 'AA', 'resname': 'GLU'},
                            {'molname': 'NA'})]}),
    (('-GLU', 'NA'),   {0: [(2,
                             3,
                            {'resname': 'GLU'},
                            {'molname': 'NA'})],
                        1: [(2,
                             4,
                            {'resname': 'GLU'},
                            {'molname': 'NA'})],
                        2: [(0,
                             5,
                            {'resname': 'GLU'},
                            {'molname': 'NA'})]}),
    (('AA-ALA', 'NA'), {0: [(0,
                             3,
                             {'molname': 'AA', 'resname': 'ALA'},
                             {'molname': 'NA'}),
                            (1,
                             4,
                             {'molname': 'AA', 'resname': 'ALA'},
                             {'molname': 'NA'})],
                        1: [(0,
                             5,
                             {'molname': 'AA', 'resname': 'ALA'},
                             {'molname': 'NA'}),
                            (1,
                             6,
                             {'molname': 'AA', 'resname': 'ALA'},
                             {'molname': 'NA'})]}),
    (('BB-ALA', 'NA'), {2: [(1,
                             3,
                             {'molname': 'BB', 'resname': 'ALA'},
                             {'molname': 'NA'})]})])
def test_init(example_system, lig_spec, expected_lig_defs):
    """
    Test if based on the ligand specs the molecules and correct nodes
    are added and annotated.
    """
    processor = polyply.src.annotate_ligands.AnnotateLigands(topology=example_system, ligands=[lig_spec])
    for mol_idx in processor.ligand_defs:
        for lig, lig_ref in zip(processor.ligand_defs[mol_idx], expected_lig_defs[mol_idx]):
            assert lig[0] == lig_ref[0]
            assert lig[1] == lig_ref[1]
            assert lig[2] == lig_ref[2]

@pytest.mark.parametrize('lig_spec', [
    ('AA#2-ALA#1', 'NA'),
    ('AA-ALA#1', '')])
def test_init_fail(example_system, lig_spec):
    """
    Test if based on the ligand specs the molecules and correct nodes
    are added and annotated.
    """
    with pytest.raises(IOError):
         polyply.src.annotate_ligands.AnnotateLigands(topology=example_system, ligands=[lig_spec])

@pytest.mark.parametrize('lig_spec, expected_mols', [
    (('AA-ALA#1', 'NA'), {0: [4],
                          1: [4]}),
    (('#0-ALA#1', 'NA'), {0: [4]}),
    (('AA#1-ALA#1', 'NA'), {1: [4]}),
    (('AA-GLU', 'NA'), {0: [4],
                        1: [4]}),
    (('AA-ALA', 'NA'), {0: [4, 5],
                        1: [4, 5]}),
    (('BB-ALA', 'NA'), {2: [4, 5]})])
def test_annotate_molecules(example_system, lig_spec, expected_mols):
    """
    Test if based on the ligand specs the molecules and correct nodes
    are added and annotated.
    """
    processor = polyply.src.annotate_ligands.AnnotateLigands(topology=example_system, ligands=[lig_spec])
    processor.run_system(example_system)
    for mol_idx, expected_nodes in expected_mols.items():
        for node in expected_nodes:
            assert "build" in example_system.molecules[mol_idx].nodes[node]
            assert "ligated" in example_system.molecules[mol_idx].nodes[node]

    for idx, mol in enumerate(example_system.molecules):
        for node in mol.nodes:
            if "ligated" in mol.nodes[node]:
                assert node in expected_mols[idx]

@pytest.mark.parametrize('lig_spec, expected_mols', [
    (('AA-ALA#1', 'NA'), {0: [4],
                          1: [4]}),
    (('#0-ALA#1', 'NA'), {0: [4]}),
    (('AA#1-ALA#1', 'NA'), {1: [4]}),
    (('AA-GLU', 'NA'), {0: [4],
                        1: [4]}),
    (('AA-ALA', 'NA'), {0: [4, 5],
                        1: [4, 5]}),
    (('BB-ALA', 'NA'), {2: [4, 5]})])
def test_split_molecules(example_system, lig_spec, expected_mols):
    """
    Test if based on the ligand specs the molecules and correct nodes
    are added and annotated.
    """
    processor = polyply.src.annotate_ligands.AnnotateLigands(topology=example_system, ligands=[lig_spec])
    processor.run_system(example_system)
    lig_count = 0
    for mol_idx, expected_nodes in expected_mols.items():
        for node in expected_nodes:
            lig_count += 1
            example_system.molecules[mol_idx].nodes[node]["position"] = np.array([1.0, 1.0, 1.0])

    processor.split_ligands()

    for mol_idx, expected_nodes in expected_mols.items():
        for node in expected_nodes:
            assert node not in example_system.molecules[mol_idx].nodes

    for idx in range(3, lig_count):
        assert "position" in example_system.molecules[idx].nodes[0]



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import (namedtuple, OrderedDict)
import json
import networkx as nx
from networkx.readwrite import json_graph
from vermouth.graph_utils import make_residue_graph
from vermouth.log_helpers import StyleAdapter, get_logger
from .polyply_parser import read_polyply
from .graph_utils import find_nodes_with_attributes

Monomer = namedtuple('Monomer', 'resname, n_blocks')
LOGGER = StyleAdapter(get_logger(__name__))

def _make_edges(force_field):
    for block in force_field.blocks.values():
        inter_types = list(block.interactions.keys())
        for inter_type in inter_types:
            if inter_type in ["bonds", "constraints" "angles"]:
               block.make_edges_from_interaction_type(type_=inter_type)

    for link in force_field.links:
        inter_types = list(link.interactions.keys())
        for inter_type in inter_types:
            if inter_type in ["bonds", "constraints", "angles"]:
                link.make_edges_from_interaction_type(type_=inter_type)

def _interpret_residue_mapping(graph, resname, new_residues):
    """
    Find all nodes corresponding to resname in graph
    and generate a corrspondance dict of these nodes
    to new resnames as defined by new_residues string
    which has the format <resname-atom1,atom2 ...>.

    Parameters:
    -----------
    graph: networkx.graph
    resname: str
    new_residues:  list[str]

    Returns:
    --------
    dict
        mapping of nodes to new residue name
    """
    atom_name_to_resname = {}
    had_atoms = []

    for new_res in new_residues:
        new_name, atoms = new_res.split("-")
        names = atoms.split(",")

        for name in names:
            if name in had_atoms:
                msg = ("You are trying to split residue {} into {} residues. "
                       "However, atom {} is mentioned more than once. This is not "
                       "allowed. ")
                raise IOError(msg.format(resname, len(new_residues), name))
            nodes = find_nodes_with_attributes(graph, resname=resname, atomname=name)
            had_atoms.append(name)

            for node in nodes:
                atom_name_to_resname[node] = new_name
    return atom_name_to_resname

class MetaMolecule(nx.Graph):
    """
    Graph that describes molecules at the residue level.
    """

    node_dict_factory = OrderedDict

    def __init__(self, *args, **kwargs):
        self.force_field = kwargs.pop('force_field', None)
        self.mol_name = kwargs.pop('mol_name', None)
        super().__init__(*args, **kwargs)
        self.molecule = None
        nx.set_node_attributes(self, True, "build")

    def add_monomer(self, current, resname, connections):
        """
        This method adds a single node and an unlimeted number
        of edges to an instance of :class::`MetaMolecule`. Note
        that matches may only refer to already existing nodes.
        But connections can be an empty list.
        """
        resids = nx.get_node_attributes(self, "resid")

        if resids:
           resid = max(resids.values()) + 1
        else:
           resid = 1

        self.add_node(current, resname=resname, resid=resid, build=True)
        for edge in connections:
            if self.has_node(edge[0]) and self.has_node(edge[1]):
                self.add_edge(edge[0], edge[1])
            else:
                msg = ("Edge {} referes to nodes that currently do"
                       "not exist. Cannot add edge to unkown nodes.")
                raise IOError(msg.format(edge))

    def get_edge_resname(self, edge):
        return [self.nodes[edge[0]]["resname"], self.nodes[edge[1]]["resname"]]

    def relabel_and_redo_res_graph(self, mapping):
        """
        Relable the nodes of `self.molecule` using `mapping`
        and regenerate the meta_molecule (i.e. residue graph).

        Parameters:
        -----------
        mapping: dict
            mapping of node-key to new residue name
        """
        # find the maximum resiude id
        max_resid = max(nx.get_node_attributes(self.molecule, "resid").values())
        # resname the residues and increase with pseudo-resid
        for node, resname in mapping.items():
            self.molecule.nodes[node]["resname"] = resname
            old_resid = self.molecule.nodes[node]["resid"]
            self.molecule.nodes[node]["resid"] = old_resid + max_resid
            self.molecule.nodes[node]["build"] = True

        # make a new residue graph and overwrite the old one
        new_meta_graph = make_residue_graph(self.molecule, attrs=('resid', 'resname'))

        # we need to do some bookkeeping for the resids
        for idx, node in enumerate(new_meta_graph.nodes):
            new_meta_graph.nodes[node]["resid"] = idx
            for atom in new_meta_graph.nodes[node]["graph"]:
                self.molecule.nodes[atom]["resid"] = idx

        self.clear()
        self.add_nodes_from(new_meta_graph.nodes(data=True))
        self.add_edges_from(new_meta_graph.edges)

    def split_residue(self, split_strings):
        """
        Split all residues defind by the in `split_strings`, which is a list
        of strings with format <resname>:<new_resname>-<atom1>,<atom2><etc>
        into new residues and also update the underlying molecule with these
        new residues.

        Parameters:
        -----------
        split_strings:  list[str]
             list of split strings

        Returns:
        dict
            mapping of the nodes to the new resnames
        """
        mapping = {}
        for split_string in split_strings:
            # split resname and new resiude definitions
            resname, *new_residues = split_string.split(":")
            # find out which atoms map to which new residues
            mapping.update(_interpret_residue_mapping(self.molecule, resname, new_residues))
        # relabel graph and redo residue graph
        self.relabel_and_redo_res_graph(mapping)
        return mapping

    @staticmethod
    def _block_graph_to_res_graph(block):
        """
        generate a residue graph from the nodes of `block`.

        Parameters
        -----------
        block: `:class:vermouth.molecule.Block`

        Returns
        -------
        :class:`nx.Graph`
        """
        res_graph = make_residue_graph(block, attrs=('resid', 'resname'))
        return res_graph

    @classmethod
    def from_monomer_seq_linear(cls, force_field, monomers, mol_name):
        """
        Constructs a meta graph for a linear molecule
        which is the default assumption from
        """

        meta_mol_graph = cls(force_field=force_field, mol_name=mol_name)
        res_count = 0

        for monomer in monomers:
            trans = 0
            while trans < monomer.n_blocks:

                if res_count != 0:
                    connect = [(res_count-1, res_count)]
                else:
                    connect = []
                trans += 1

                meta_mol_graph.add_monomer(res_count, monomer.resname, connect)
                res_count += 1
        return meta_mol_graph

    @classmethod
    def from_json(cls, force_field, json_file, mol_name):
        """
        Constructs a :class::`MetaMolecule` from a json file
        using the networkx json package.
        """
        with open(json_file) as file_:
            data = json.load(file_)

        init_graph = nx.Graph(json_graph.node_link_graph(data))
        graph = nx.Graph(node_dict_factory=OrderedDict)
        nodes = list(init_graph.nodes)
        nodes.sort()

        for node in nodes:
            attrs = init_graph.nodes[node]
            if "resid" not in attrs:
                LOGGER.warning("Node {} has no resid. Setting resid to {} + 1.", node, node)
                try:
                    attrs["resid"] = node + 1
                except TypeError:
                    msg = "Couldn't add 1 to node. Either provide resids or use integers as node keys."
                    raise IOError(msg)
            graph.add_node(node, **attrs)

        graph.add_edges_from(init_graph.edges(data=True))
        meta_mol = cls(graph, force_field=force_field, mol_name=mol_name)
        return meta_mol

    @classmethod
    def from_itp(cls, force_field, itp_file, mol_name):
        """
        Constructs a :class::`MetaMolecule` from an itp file.
        """
        with open(itp_file) as file_:
            lines = file_.readlines()
            read_polyply(lines, force_field)

        graph = MetaMolecule._block_graph_to_res_graph(force_field.blocks[mol_name])
        meta_mol = cls(graph, force_field=force_field, mol_name=mol_name)
        meta_mol.molecule = force_field.blocks[mol_name].to_molecule()
        return meta_mol

    #TODO Add from_molecule method
    @classmethod
    def from_block(cls, force_field, mol_name):
        """
        Constructs a :class::`MetaMolecule` from an vermouth.molecule.
        """
        _make_edges(force_field)
        block = force_field[mol_name]
        graph = MetaMolecule._block_graph_to_res_graph(block)
        meta_mol = cls(graph, force_field=force_field, mol_name=mol_name)
        meta_mol.molecule = force_field.blocks[mol_name].to_molecule()
        return meta_mol



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx as nx
import numpy as np
import scipy
import vermouth
from vermouth.log_helpers import StyleAdapter, get_logger
from .minimizer import optimize_geometry
from .processor import Processor
from .linalg_functions import (u_vect, center_of_geometry,
                               radius_of_gyration)
from .topology import replace_defined_interaction
"""
Processor generating coordinates for all residues of a meta_molecule
matching those in the meta_molecule.molecule attribute.
"""

LOGGER = StyleAdapter(get_logger(__name__))

def find_atoms(molecule, attr, value):
    """
    Find all nodes of a `vermouth.molecule.Molecule` that have the
    attribute `attr` with the corresponding value of `value`.

    Parameters
    ----------
    molecule: :class:vermouth.molecule.Molecule
    attr: str
         attribute that a node needs to have
    value:
         corresponding value

    Returns
    ----------
    list
       list of nodes found
    """
    nodes = []
    for node in molecule.nodes:
        if attr in molecule.nodes[node] and molecule.nodes[node][attr] == value:
            nodes.append(node)

    return nodes

def find_interaction_involving(block, current_node, prev_node):
    """
    Given a list of `interactions` in vermouth format, find an
    interaction from bonds, constraints, or virtual-sites that
    involves the `current_node` and `prev_node`. Return if this
    interaction is a virtual-site and the corresponding parameter.

    Parameters
    -----------
    block:  :class:`vermouth.molecule.Block`
         vermouth block
    current_node:   int
         node index
    prev_node:      int
         node index

    Returns:
    ---------
    bool
      is the interaction a virtual-site
    vermouth.Interaction
      interaction definition
    str
      interaction type
    """
    interactions = block.interactions
    for inter_type in ["bonds", "constraints", "virtual_sitesn",
                       "virtual_sites2", "virtual_sites3", "virtual_sites4"]:
        inters = interactions.get(inter_type, [])
        for interaction in inters:
            if current_node in interaction.atoms:
                if prev_node in interaction.atoms and inter_type in ["bonds", "constraints"]:
                    return False, interaction, inter_type
                elif prev_node in interaction.atoms and inter_type.split("_")[0] == "virtual":
                    return True, interaction, inter_type
    else:
        msg = '''Cannot build template for residue {}. No constraint, bond, virtual-site
                 linking atom {} and atom {}.'''
        raise IOError(msg.format(block.nodes[0]["resname"], prev_node, current_node))

def _expand_inital_coords(block, bond=None, pos=None, fixed=None,
                          iterations=50, weight="weight", max_box=1.0):
    """
    Given a `graph` generate initial coordinates in three dimensions
    using the Kamada-Kawai algorithm.

    Parameters
    -----------
    block:   networkx.Graph

    Returns
    ---------
    dict
      dictonary of node index and position
    """
    return nx.kamada_kawai_layout(block, dim=3)

def compute_volume(molecule, block, coords, nonbond_params, treshold=1e-18):
    """
    Given a `block`, which is part of `molecule` and
    has the coordinates `coord` compute the radius
    of gyration taking into account the volume of each
    particle. The volume of a particle is considered to be
    the sigma value of it's LJ self interaction parameter.

    Parameters
    ----------
    molecule:  :class:vermouth.molecule.Molecule
    block:     :class:vermouth.molecule.Block
    coords:    :class:dict
        dictionary of positions in from node_idx: np.array
    treshold: float
        distance from center of geometry at which the
        particle is not taken into account for the volume
        computation

    Returns
    -------
    radius of gyration
    """
    n_atoms = len(coords)
    points = np.array(list(coords.values()))
    res_center_of_geometry = center_of_geometry(points)
    geom_vects = np.zeros((n_atoms, 3))
    idx = 0

    for node, coord in coords.items():
        atom_key = block.nodes[node]["atype"]
        rad = float(nonbond_params[frozenset([atom_key, atom_key])]["nb1"])
        diff = coord - res_center_of_geometry
        if np.linalg.norm(diff) < treshold:
           continue
        else:
           geom_vects[idx, :] = diff + u_vect(diff) * rad
        idx += 1

    if geom_vects.shape[0] > 1:
        radgyr = radius_of_gyration(geom_vects)
    else:
        radgyr = rad

    return radgyr

def map_from_CoG(coords):
    """
    Compute the center of geometry
    of `coords` and return each position
    as vector between the center of geometry
    and the original positon.

    Parameters
    ----------
    coords:   :class:dict
        dictionary of coordinates

    Returns
    --------
    dict
     dictionary of node idx and CoG vector
    """
    points = np.array(list(coords.values()))
    res_center_of_geometry = center_of_geometry(points)
    out_vectors = {}
    for key, coord in coords.items():
        diff = coord - res_center_of_geometry
        out_vectors[key] = diff

    return out_vectors

def _relabel_interaction_atoms(interaction, mapping):
    """
    Relables the atoms in interaction according to the
    rules defined in mapping.

    Parameters
    ----------
    interaction: `vermouth.molecule.Interaction`
    mapping: `:class:dict`

    Returns
    -------
    interaction: `vermouth.molecule.Interaction`
        the new interaction with updated atoms
    """
    new_atoms = [mapping[atom] for atom in interaction.atoms]
    new_interaction = interaction._replace(atoms=new_atoms)
    return new_interaction

def extract_block(molecule, resname, defines):
    """
    Given a `vermouth.molecule` and a `resname`
    extract the information of a block from the
    molecule definition and replace all defines
    if any are found.

    Parameters
    ----------
    molecule:  :class:vermouth.molecule.Molecule
    resname:   str
    defines:   dict
      dict of type define: value

    Returns
    -------
    :class:vermouth.molecule.Block
    """
    nodes = find_atoms(molecule, "resname", resname)
    resid = molecule.nodes[nodes[0]]["resid"]
    block = vermouth.molecule.Block()

    # select all nodes with the same first resid and
    # make sure the block node labels are atomnames
    # also build a correspondance dict between node
    # label in the molecule and in the block for
    # relabeling the interactions
    mapping = {}
    for node in nodes:
        attr_dict = molecule.nodes[node]
        if attr_dict["resid"] == resid:
            block.add_node(attr_dict["atomname"], **attr_dict)
            mapping[node] = attr_dict["atomname"]

    for inter_type in molecule.interactions:
        for interaction in molecule.interactions[inter_type]:
            if all(atom in mapping for atom in interaction.atoms):
                interaction = replace_defined_interaction(interaction, defines)
                interaction = _relabel_interaction_atoms(interaction, mapping)
                block.interactions[inter_type].append(interaction)

    for inter_type in ["bonds", "constraints", "virtual_sitesn",
                       "virtual_sites2", "virtual_sites3", "virtual_sites4"]:
        block.make_edges_from_interaction_type(inter_type)

    if not nx.is_connected(block):
        msg = ('\n Residue {} with id {} consistes of two disconnected parts. '
               'Make sure all atoms/particles in a residue are connected by bonds,'
               ' constraints or virual-sites.')
        raise IOError(msg.format(resname, resid))

    return block

class GenerateTemplates(Processor):
    """
    This processor takes a a class:`polyply.src.MetaMolecule` and
    creates a block for each unique residue type in the molecule
    as well as positions for that block. These blocks are stored
    in the templates attribute. The processor also stores the volume
    of each block in the volume attribute.
    """
    def __init__(self, topology, max_opt, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.templates = {}
        self.resnames = []
        self.volumes = {}
        self.max_opt = max_opt
        self.topology = topology

    def _gen_templates(self, meta_molecule):
        """
        Generate blocks for each unique residue by extracting the
        block information, placing initial coordinates, and geometry
        optimizing those coordinates. Subsequently compute volume

        Parameters
        ----------
        meta_molecule: :class:`polyply.src.meta_molecule.MetaMolecule`

        Returns
        ---------
        templates:  dict
           dict of resname: block
        volumes:    dict
           dict of name: volume
        """
        resnames = set(nx.get_node_attributes(meta_molecule.molecule,
                                              "resname").values())
        templates = {}
        volumes = {}

        for resname in resnames:
            if not resname in self.resnames:
                self.resnames.append(resname)

                block = extract_block(meta_molecule.molecule, resname,
                                      self.topology.defines)
                opt_counter = 0
                while True:
                    coords = _expand_inital_coords(block)
                    success, coords = optimize_geometry(block, coords, ["bonds", "constraints"])
                    success, coords = optimize_geometry(block, coords, ["bonds", "constraints", "angles"])
                    success, coords = optimize_geometry(block, coords, ["bonds", "constraints", "angles", "dihedrals"])

                    if success:
                        break
                    elif opt_counter > self.max_opt:
                        LOGGER.warning("Failed to optimize structure for block {}.", resname)
                        LOGGER.warning("Proceeding with unoptimized coordinates.")
                        LOGGER.warning("Usually this is OK, but check your final structure.""")
                        break
                    else:
                        opt_counter += 1

                self.volumes[resname] = compute_volume(meta_molecule, block, coords, self.topology.nonbond_params)
                coords = map_from_CoG(coords)
                self.templates[resname] = coords

        return templates, volumes

    def run_molecule(self, meta_molecule):
        """
        Execute the generation of templates and set the template
        and volume attribute.
        """
        templates, volumes = self._gen_templates(meta_molecule)
        meta_molecule.templates = self.templates
        self.topology.volumes = self.volumes
        return meta_molecule



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
import numpy as np
from vermouth.parser_utils import SectionLineParser

class BuildDirector(SectionLineParser):

    COMMENT_CHAR = ';'

    def __init__(self, molecules):
        super().__init__()
        self.molecules = molecules
        self.current_molidxs = []
        self.build_options = defaultdict(list)
        self.current_molname = None
        self.rw_options = dict()

    @SectionLineParser.section_parser('molecule')
    def _molecule(self, line, lineno=0):
        """
        Parses the lines in the '[molecule]'
        directive and stores it.
        """
        tokens = line.split()
        self.current_molname = tokens[0]
        self.current_molidxs = np.arange(float(tokens[1]), float(tokens[2]), 1., dtype=int)

    @SectionLineParser.section_parser('molecule', 'cylinder', geom_type="cylinder")
    @SectionLineParser.section_parser('molecule', 'sphere', geom_type="sphere")
    @SectionLineParser.section_parser('molecule', 'rectangle', geom_type="rectangle")
    def _parse_geometry(self, line, lineno, geom_type):
        """
        Parses the lines in the '[<geom_type>]'
        directive and stores it. The format of the directive
        is as follows:

        <resname><resid_start><resid_stop><x><y><z><parameters>

        parameters depends on the geometry type and contains
        for:

        spheres - radius
        cylinder - radius and 1/2 d
        reactangle - 1/2 a, b, c which are the lengths
        """
        tokens = line.split()
        geometry_def = self._base_parser_geometry(tokens, geom_type)
        for idx in self.current_molidxs:
            self.build_options[(self.current_molname, idx)].append(geometry_def)

    @SectionLineParser.section_parser('molecule', 'rw_restriction')
    def _rw_restriction(self, line, lineno=0):
        """
        Restrict random walk in specific direction.
        """
        tokens = line.split()
        geometry_def = {"resname": tokens[0],
                        "start": int(tokens[1]),
                        "stop":  int(tokens[2]),
                        "parameters": [np.array([float(tokens[3]), float(tokens[4]), float(tokens[5])]),
                                       float(tokens[6])]}
        for idx in self.current_molidxs:
            self.rw_options[(self.current_molname, idx)] = geometry_def

    @SectionLineParser.section_parser('molecule', 'chiral')
    def _chiral(self, line, lineno=0):
        """
        Parses the lines in the '[chiral]'
        directive and stores it.
        """
        pass

    @SectionLineParser.section_parser('molecule', 'isomer')
    def _isomer(self, line, lineno=0):
        """
        Parses the lines in the '[isomer]'
        directive and stores it.
        """
        pass

    def finalize(self, lineno=0):
        """
        Tag each molecule node with the chirality and build options
        if that molecule is mentioned in the build file by name.
        """
        for mol_idx, molecule in enumerate(self.molecules):
            if (molecule.mol_name, mol_idx) in self.build_options:
                for option in self.build_options[(molecule.mol_name, mol_idx)]:
                    self._tag_nodes(molecule, "restraints", option)

            if (molecule.mol_name, mol_idx)  in self.rw_options:
                self._tag_nodes(molecule, "rw_options",
                                self.rw_options[(molecule.mol_name, mol_idx)])

        super().finalize

    @staticmethod
    def _tag_nodes(molecule, keyword, option):
        resids = np.arange(option['start'], option['stop'], 1.)
        resname = option["resname"]
        for node in molecule.nodes:
            if molecule.nodes[node]["resid"] in resids\
            and molecule.nodes[node]["resname"] == resname:
                molecule.nodes[node][keyword] = molecule.nodes[node].get(keyword, []) +\
                                                [option['parameters']]

    @staticmethod
    def _base_parser_geometry(tokens, _type):
        geometry_def = {}
        geometry_def["resname"] = tokens[0]
        geometry_def["start"] = float(tokens[1])
        geometry_def["stop"] = float(tokens[2])

        point = np.array([float(tokens[4]), float(tokens[5]), float(tokens[6])])
        parameters = [tokens[3], point]

        for param in tokens[7:]:
            parameters.append(float(param))

        parameters.append(_type)
        geometry_def["parameters"] = parameters
        return geometry_def

def read_build_file(lines, molecules):
    """
    Parses `lines` of itp format and adds the
    molecule as a block to `force_field`.

    Parameters
    ----------
    lines: list
        list of lines of an itp file
    force_field: :class:`vermouth.forcefield.ForceField`
    """
    director = BuildDirector(molecules)
    return list(director.parse(iter(lines)))
